import glob
import os

from esdl.profiles.excelprofilemanager import ExcelProfileManager
from esdl.profiles.influxdbprofilemanager import ConnectionSettings
from esdl.profiles.influxdbprofilemanager import InfluxDBProfileManager

"""
Place raw excel files in fodler: raw_data_files_folder
Database definition:

Database name: input esdl id
Measurement: asset name
Fields: Variables per asset
Tags/fiters: {tag: "output_esdl_id, value: id}
"""

# Settings for the reading excel files and creating a influxDB
# raw_data_files_folder = "C:\\Projects_gitlab\\database_files_test\\"
# input_energy_system_id = "15174819-d1af-4ba6-9f1d-2cd07991f14a"
# output_energy_system_id = "a33fe8db-8bdb-45a0-b1e7-69c348001672"
# influxdb_conn_settings = ConnectionSettings(
#     host="localhost",
#     port=8086,
#     username=None,
#     password=None,
#     database=input_energy_system_id,
#     ssl=False,
#     verify_ssl=False,
# )
# optim_simulation_tag = {"output_esdl_id": output_energy_system_id}

# Temp code below for electrolyzer
# Settings for the reading excel files and creating a influxDB
# raw_data_files_folder = (
#     "C:\\Projects_gitlab\\NWN_dev\\rtc-tools-milp-network\\tests\\models\\unit_cases_electricity\\"
#     "electrolyzer\\input\\Dummy_values\\"
# )
raw_data_files_folder = (
    "C:\\Projects_gitlab\\NWN_dev\\rtc-tools-milp-network\\tests\\models\\unit_cases_electricity\\"
    "electrolyzer\\input\\Profiles\\"
)
# input_energy_system_id = "15174819-d1af-4ba6-9f1d-2cd07991f14a"
# output_energy_system_id = "a33fe8db-8bdb-45a0-b1e7-69c348001672"
influxdb_conn_settings = ConnectionSettings(
    host="omotes-poc-test.hesi.energy",
    port=8086,
    username="write-user",
    password="nwn_write_test",
    database="multicommodity_test",
    ssl=False,
    verify_ssl=False,
)
optim_simulation_tag = {"output_esdl_id": "1"}
# optim_simulation_tag = {"output_esdl_id": output_energy_system_id}

excel_files = glob.glob(os.path.join(raw_data_files_folder, "*.xlsx"))


for file in excel_files:
    print("Read data from Excel")
    excel_prof_read = ExcelProfileManager()
    excel_prof_read.load_excel(file)

    print("Create database")
    influxdb_profile_manager_create_new = InfluxDBProfileManager(
        influxdb_conn_settings, excel_prof_read
    )
    asset_name = file.split("\\")[-1].replace(".xlsx", "")
    _ = influxdb_profile_manager_create_new.save_influxdb(
        measurement=asset_name,
        field_names=influxdb_profile_manager_create_new.profile_header[1:],
        tags=optim_simulation_tag,
    )



import locale
import logging
import os
import time

from mesido.esdl.esdl_additional_vars_mixin import ESDLAdditionalVarsMixin
from mesido.esdl.esdl_mixin import ESDLMixin
from mesido.head_loss_class import HeadLossOption
from mesido.techno_economic_mixin import TechnoEconomicMixin
from mesido.workflows.goals.minimize_tco_goal import MinimizeTCO
from mesido.workflows.io.write_output import ScenarioOutput
from mesido.workflows.utils.adapt_profiles import (
    adapt_hourly_year_profile_to_day_averaged_with_hourly_peak_day,
)
from mesido.workflows.utils.helpers import main_decorator

import numpy as np

from rtctools._internal.alias_tools import AliasDict
from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.goal_programming_mixin_base import Goal
from rtctools.optimization.linearized_order_goal_programming_mixin import (
    LinearizedOrderGoalProgrammingMixin,
)
from rtctools.optimization.single_pass_goal_programming_mixin import (
    CachingQPSol,
    SinglePassGoalProgrammingMixin,
)
from rtctools.util import run_optimization_problem


DB_HOST = "172.17.0.2"
DB_PORT = 8086
DB_NAME = "Warmtenetten"
DB_USER = "admin"
DB_PASSWORD = "admin"

logger = logging.getLogger("WarmingUP-MPC")
logger.setLevel(logging.INFO)

locale.setlocale(locale.LC_ALL, "")

ns = {"fews": "http://www.wldelft.nl/fews", "pi": "http://www.wldelft.nl/fews/PI"}

WATT_TO_MEGA_WATT = 1.0e6
WATT_TO_KILO_WATT = 1.0e3


class TargetHeatGoal(Goal):
    priority = 1

    order = 2

    def __init__(self, state, target):
        self.state = state

        self.target_min = target
        self.target_max = target
        try:
            self.function_range = (-1.0e6, max(2.0 * max(target.values), 1.0e6))
            self.function_nominal = max(np.median(target.values), 1.0e6)
        except Exception:
            self.function_range = (-1.0e6, max(2.0 * target, 1.0e6))
            self.function_nominal = max(target, 1.0e6)

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(self.state)


class SolverHIGHS:
    def solver_options(self):
        options = super().solver_options()
        options["casadi_solver"] = self._qpsol
        options["solver"] = "highs"
        highs_options = options["highs"] = {}
        if hasattr(self, "_stage"):
            if self._stage == 1:
                highs_options["mip_rel_gap"] = 0.005
            else:
                highs_options["mip_rel_gap"] = 0.02
        else:
            highs_options["mip_rel_gap"] = 0.02

        options["gurobi"] = None

        return options


class SolverGurobi:
    def solver_options(self):
        options = super().solver_options()
        options["casadi_solver"] = self._qpsol
        options["solver"] = "gurobi"
        gurobi_options = options["gurobi"] = {}
        if hasattr(self, "_stage"):
            if self._stage == 1:
                gurobi_options["MIPgap"] = 0.005
            else:
                gurobi_options["MIPgap"] = 0.02
        gurobi_options["MIPgap"] = 0.02
        gurobi_options["threads"] = 4
        gurobi_options["LPWarmStart"] = 2

        options["highs"] = None

        return options


class EndScenarioSizing(
    SolverHIGHS,
    ScenarioOutput,
    ESDLAdditionalVarsMixin,
    TechnoEconomicMixin,
    LinearizedOrderGoalProgrammingMixin,
    SinglePassGoalProgrammingMixin,
    ESDLMixin,
    CollocatedIntegratedOptimizationProblem,
):
    """
    This class is the base class to run all the other EndScenarioSizing classes from.

    HIGHS is now the standard solver and gurobi only to be used when called specifically.

    Goal priorities are:
    1. Demand matching (e.g. minimize (heat demand - heat consumed))
    2. minimize TCO = Capex + Opex*lifetime
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.heat_network_settings["minimum_velocity"] = 0.0  # 0.001
        self.heat_network_settings["maximum_velocity"] = 3.0
        self.heat_network_settings["head_loss_option"] = HeadLossOption.NO_HEADLOSS

        self._override_hn_options = {}

        self._number_of_years = 30.0

        self.__indx_max_peak = None
        self.__day_steps = 5

        # self._override_pipe_classes = {}

        # variables for solver settings
        self._qpsol = None

        self._hot_start = False

        # Store (time taken, success, objective values, solver stats) per priority
        self._priorities_output = []
        self.__priority = None
        self.__priority_timer = None

        self.__heat_demand_bounds = dict()
        self.__heat_demand_nominal = dict()

        self._save_json = False

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)
        parameters["peak_day_index"] = self.__indx_max_peak
        parameters["time_step_days"] = self.__day_steps
        parameters["number_of_years"] = self._number_of_years
        return parameters

    def pre(self):
        self._qpsol = CachingQPSol()

        super().pre()

    def read(self):
        """
        Reads the yearly profile with hourly time steps and adapt to a daily averaged profile
        except for the day with the peak demand.
        """
        super().read()

        (
            self.__indx_max_peak,
            self.__heat_demand_nominal,
        ) = adapt_hourly_year_profile_to_day_averaged_with_hourly_peak_day(self, self.__day_steps)

        logger.info("HeatProblem read")

    def bounds(self):
        bounds = super().bounds()
        bounds.update(self.__heat_demand_bounds)
        return bounds

    def variable_nominal(self, variable):
        try:
            return self.__heat_demand_nominal[variable]
        except KeyError:
            return super().variable_nominal(variable)

    def energy_system_options(self):
        # TODO: make empty placeholder in HeatProblem we don't know yet how to put the global
        #  constraints in the ESDL e.g. min max pressure
        options = super().energy_system_options()
        options["maximum_temperature_der"] = np.inf
        options["heat_loss_disconnected_pipe"] = True
        # options.update(self._override_hn_options)
        return options

    def path_goals(self):
        goals = super().path_goals().copy()
        bounds = self.bounds()

        for demand in self.energy_system_components["heat_demand"]:
            # target = self.get_timeseries(f"{demand}.target_heat_demand_peak")
            target = self.get_timeseries(f"{demand}.target_heat_demand")
            if bounds[f"{demand}.HeatIn.Heat"][1] < max(target.values):
                logger.warning(
                    f"{demand} has a flow limit, {bounds[f'{demand}.HeatIn.Heat'][1]}, "
                    f"lower that wat is required for the maximum demand {max(target.values)}"
                )
            state = f"{demand}.Heat_demand"

            goals.append(TargetHeatGoal(state, target))
        return goals

    def goals(self):
        goals = super().goals().copy()
        # We do a minization of TCO consisting of CAPEX and OPEX over 25 years
        # CAPEX is based upon the boolean placement variables and the optimized maximum sizes
        # Note that CAPEX for geothermal and ATES is also dependent on the amount of doublets
        # In practice this means that the CAPEX is mainly driven by the peak day problem
        # The OPEX is based on the Source strategy which is computed on the __daily_avg variables
        # The OPEX thus is based on an avg strategy and discrepancies due to fluctuations intra-day
        # are possible.
        # The idea behind the two timelines is that the optimizer can make the OPEX vs CAPEX
        # trade-offs

        goals.append(MinimizeTCO(priority=2, number_of_years=self._number_of_years))

        return goals

    def constraints(self, ensemble_member):
        constraints = super().constraints(ensemble_member)

        for a in self.energy_system_components.get("ates", []):
            stored_heat = self.state_vector(f"{a}.Stored_heat")
            constraints.append(((stored_heat[-1] - stored_heat[0]), 0.0, np.inf))
            ates_temperature = self.state_vector(f"{a}.Temperature_ates")
            constraints.append(((ates_temperature[-1] - ates_temperature[0]), 0.0, np.inf))

        for b in self.energy_system_components.get("heat_buffer", {}):
            vars = self.state_vector(f"{b}.Heat_buffer")
            symbol_stored_heat = self.state_vector(f"{b}.Stored_heat")
            constraints.append((symbol_stored_heat[self.__indx_max_peak], 0.0, 0.0))
            for i in range(len(self.times())):
                if i < self.__indx_max_peak or i > (self.__indx_max_peak + 23):
                    constraints.append((vars[i], 0.0, 0.0))

        return constraints

    def history(self, ensemble_member):
        return AliasDict(self.alias_relation)

    def __state_vector_scaled(self, variable, ensemble_member):
        canonical, sign = self.alias_relation.canonical_signed(variable)
        return (
            self.state_vector(canonical, ensemble_member) * self.variable_nominal(canonical) * sign
        )

    def solver_options(self):
        options = super().solver_options()
        if options["solver"] == "highs":
            highs_options = options["highs"]
            if self.__priority == 1:
                highs_options["time_limit"] = 100
            else:
                highs_options["time_limit"] = 100000
        return options

    def solver_success(self, solver_stats, log_solver_failure_as_error):
        success, log_level = super().solver_success(solver_stats, log_solver_failure_as_error)

        # Allow time-outs for CPLEX and CBC
        if (
            solver_stats["return_status"] == "time limit exceeded"
            or solver_stats["return_status"] == "stopped - on maxnodes, maxsols, maxtime"
        ):
            if self.objective_value > 1e10:
                # Quick check on the objective value. If no solution was
                # found, this is typically something like 1E50.
                return success, log_level

            return True, logging.INFO
        else:
            return success, log_level

    def priority_started(self, priority):
        goals_print = set()
        for goal in [*self.path_goals(), *self.goals()]:
            if goal.priority == priority:
                goals_print.update([str(type(goal))])
        logger.info(f"{goals_print}")
        self.__priority = priority
        self.__priority_timer = time.time()

        super().priority_started(priority)

    def priority_completed(self, priority):
        super().priority_completed(priority)

        self._hot_start = True

        time_taken = time.time() - self.__priority_timer
        self._priorities_output.append(
            (
                priority,
                time_taken,
                True,
                self.objective_value,
                self.solver_stats,
            )
        )
        if priority == 1 and self.objective_value > 1e-6:
            raise RuntimeError("The heating demand is not matched")

    def post(self):
        # In case the solver fails, we do not get in priority_completed(). We
        # append this last priority's statistics here in post().
        # TODO: check if we still need this small part of code below
        success, _ = self.solver_success(self.solver_stats, False)
        if not success:
            time_taken = time.time() - self.__priority_timer
            self._priorities_output.append(
                (
                    self.__priority,
                    time_taken,
                    False,
                    self.objective_value,
                    self.solver_stats,
                )
            )

        super().post()
        results = self.extract_results()
        parameters = self.parameters(0)
        # bounds = self.bounds()
        # Optimized ESDL
        self._write_updated_esdl(self._ESDLMixin__energy_system_handler.energy_system)

        for d in self.energy_system_components.get("heat_demand", []):
            realized_demand = results[f"{d}.Heat_demand"]
            target = self.get_timeseries(f"{d}.target_heat_demand").values
            timesteps = np.diff(self.get_timeseries(f"{d}.target_heat_demand").times)
            parameters[f"{d}.target_heat_demand"] = target.tolist()
            delta_energy = np.sum((realized_demand - target)[1:] * timesteps / 1.0e9)
            if delta_energy >= 1.0:
                logger.warning(f"For demand {d} the target is not matched by {delta_energy} GJ")

        if os.path.exists(self.output_folder) and self._save_json:
            self._write_json_output()


class EndScenarioSizingHIGHS(EndScenarioSizing):
    """
    HIGHS is now the standard solver and gurobi only to be used when called specifically.
    Currently, the classes in HIGHS are maintained such that the same 'old' function calling can be
    used for the code running in NWN.
    """

    pass


class EndScenarioSizingGurobi(SolverGurobi, EndScenarioSizing):
    """
    Uses Gurobi as the solver for the EndScenarioSizing problem.
    """

    pass


class EndScenarioSizingDiscounted(EndScenarioSizing):
    """
    The discounted annualized is utilised as the objective function.
    The change of the objective function is done by changing the option 'discounted_annulized_cost'
    to True

    Goal priorities are:
    1. Match heat demand with target
    2. minimize TCO = Anualized capex (function of technical lifetime of individual assets) +
    Opex*timehorizon
    """

    def energy_system_options(self):
        options = super().energy_system_options()

        options["discounted_annualized_cost"] = True

        return options


class EndScenarioSizingDiscountedHIGHS(EndScenarioSizingDiscounted):
    pass


class EndScenarioSizingDiscountedGurobi(SolverGurobi, EndScenarioSizingDiscounted):
    pass


class EndScenarioSizingHeadLoss(EndScenarioSizing):
    """
    EndScenarioSizing optimisation including the linearised inequality DarcyWeisbach Head loss
    relations

    The minimize_head_losses setting can be set to False as the pumping costs can be included in
    the TCO, it does require a price profile to be added to the electricity carrier, else it should
     be set to True.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.heat_network_settings["head_loss_option"] = (
            HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
        )
        self.heat_network_settings["minimize_head_losses"] = True


class EndScenarioSizingHeadLossDiscounted(EndScenarioSizingHeadLoss, EndScenarioSizingDiscounted):
    pass


class SettingsStaged:
    """
    Additional settings to be used when a staged approach should be implemented.
    Staged approach currently entails 2 stages:
    1. optimisation without heat losses and thus a much smaller MIPgap (in solver options) is used
    to ensure the bounds set for the second stage are not limiting the optimal solution
    2. optimisation including heat losses with updated boolean bounds (smaller range) of asset
    sizes and flow directions.
    """

    _stage = 0

    def __init__(
        self,
        stage=None,
        boolean_bounds: list = None,
        priorities_output: list = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self._stage = stage
        self.__boolean_bounds = boolean_bounds

        if self._stage == 1:
            self.heat_network_settings["minimum_velocity"] = 0.0
            self.heat_network_settings["head_loss_option"] = HeadLossOption.NO_HEADLOSS
            self.heat_network_settings["minimize_head_losses"] = False

        if self._stage == 2 and priorities_output:
            self._priorities_output = priorities_output

    def energy_system_options(self):
        options = super().energy_system_options()
        if self._stage == 1:
            options["neglect_pipe_heat_losses"] = True
            self.heat_network_settings["minimum_velocity"] = 0.0

        return options

    def bounds(self):
        bounds = super().bounds()

        if self._stage == 2:
            bounds.update(self.__boolean_bounds)

        return bounds


class EndScenarioSizingStaged(SettingsStaged, EndScenarioSizing):
    pass


class EndScenarioSizingStagedHIGHS(EndScenarioSizingStaged):
    pass


class EndScenarioSizingStagedGurobi(SolverGurobi, EndScenarioSizingStaged):
    pass


class EndScenarioSizingDiscountedStaged(SettingsStaged, EndScenarioSizingDiscounted):
    pass


class EndScenarioSizingDiscountedStagedHIGHS(EndScenarioSizingDiscountedStaged):
    pass


class EndScenarioSizingDiscountedStagedGurobi(SolverGurobi, EndScenarioSizingDiscountedStaged):
    pass


class EndScenarioSizingHeadLossStaged(SettingsStaged, EndScenarioSizingHeadLoss):
    pass


class EndScenarioSizingHeadLossStagedGurobi(SolverGurobi, EndScenarioSizingHeadLossStaged):
    pass


class EndScenarioSizingHeadLossDiscountedStaged(
    SettingsStaged, EndScenarioSizingHeadLossDiscounted
):
    pass


class EndScenarioSizingHeadLossDiscountedStagedGurobi(
    SolverGurobi, EndScenarioSizingHeadLossDiscountedStaged
):
    pass


def run_end_scenario_sizing_no_heat_losses(
    end_scenario_problem_class,
    **kwargs,
):
    """
    This function is used to run end_scenario_sizing problem without milp losses. This is a
    simplification from the fully staged approach allowing users to more quickly iterate over
    results.

    Parameters
    ----------
    end_scenario_problem_class : The end scenario problem class.
    staged_pipe_optimization : Boolean to toggle between the staged or non-staged approach

    Returns
    -------

    """
    import time

    assert issubclass(
        end_scenario_problem_class, SettingsStaged
    ), "A staged problem class is required as input for the sizing without heat_losses"

    start_time = time.time()
    solution = run_optimization_problem(
        end_scenario_problem_class,
        stage=1,
        **kwargs,
    )

    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))

    return solution


def run_end_scenario_sizing(
    end_scenario_problem_class,
    staged_pipe_optimization=True,
    **kwargs,
):
    """
    This function is used to run end_scenario_sizing problem. There are a few variations of the
    same basic class. The main functionality this function adds is the staged approach, where
    we first solve without heat_losses, to then solve the same problem with milp losses but
    constraining the problem to only allow for the earlier found pipe classes and one size up.

    This staged approach is done to speed up the problem, as the problem without milp losses is
    much faster as it avoids inequality big_m constraints for the milp to discharge on pipes. The
    one size up possibility is to avoid infeasibilities in compensating for the milp losses.

    Parameters
    ----------
    end_scenario_problem_class : The end scenario problem class.
    staged_pipe_optimization : Boolean to toggle between the staged or non-staged approach

    Returns
    -------

    """
    import time

    boolean_bounds = {}
    priorities_output = []

    start_time = time.time()
    if staged_pipe_optimization and issubclass(end_scenario_problem_class, SettingsStaged):
        solution = run_optimization_problem(
            end_scenario_problem_class,
            stage=1,
            **kwargs,
        )
        results = solution.extract_results()
        parameters = solution.parameters(0)
        bounds = solution.bounds()

        # We give bounds for stage 2 by allowing one DN sizes larger than what was found in the
        # stage 1 optimization.
        pc_map = solution.get_pipe_class_map()
        for pipe_classes in pc_map.values():
            v_prev = 0.0
            first_pipe_class = True
            for var_name in pipe_classes.values():
                v = results[var_name][0]
                if first_pipe_class and abs(v) == 1.0:
                    boolean_bounds[var_name] = (abs(v), abs(v))
                elif abs(v) == 1.0:
                    boolean_bounds[var_name] = (0.0, abs(v))
                elif v_prev == 1.0:
                    boolean_bounds[var_name] = (0.0, 1.0)
                else:
                    boolean_bounds[var_name] = (abs(v), abs(v))
                v_prev = v
                first_pipe_class = False

        for asset in [
            *solution.energy_system_components.get("heat_source", []),
            *solution.energy_system_components.get("heat_buffer", []),
        ]:
            var_name = f"{asset}_aggregation_count"
            lb = results[var_name][0]
            ub = solution.bounds()[var_name][1]
            if round(lb) >= 1:
                boolean_bounds[var_name] = (lb, ub)

        t = solution.times()
        from rtctools.optimization.timeseries import Timeseries

        for p in solution.energy_system_components.get("heat_pipe", []):
            if p in solution.hot_pipes and parameters[f"{p}.area"] > 0.0:
                lb = []
                ub = []
                bounds_pipe = bounds[f"{p}__flow_direct_var"]
                for i in range(len(t)):
                    r = results[f"{p}__flow_direct_var"][i]
                    # bound to roughly represent 4km of milp losses in pipes
                    lb.append(
                        r
                        if abs(results[f"{p}.Q"][i] / parameters[f"{p}.area"]) > 2.5e-2
                        else bounds_pipe[0]
                    )
                    ub.append(
                        r
                        if abs(results[f"{p}.Q"][i] / parameters[f"{p}.area"]) > 2.5e-2
                        else bounds_pipe[1]
                    )

                boolean_bounds[f"{p}__flow_direct_var"] = (Timeseries(t, lb), Timeseries(t, ub))
                try:
                    r = results[f"{p}__is_disconnected"]
                    boolean_bounds[f"{p}__is_disconnected"] = (Timeseries(t, r), Timeseries(t, r))
                except KeyError:
                    pass
        priorities_output = solution._priorities_output

    solution = run_optimization_problem(
        end_scenario_problem_class,
        stage=2,
        boolean_bounds=boolean_bounds,
        priorities_output=priorities_output,
        **kwargs,
    )

    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))

    return solution


@main_decorator
def main(runinfo_path, log_level):
    logger.info("Run Scenario Sizing")

    kwargs = {
        "write_result_db_profiles": False,
        "influxdb_host": "localhost",
        "influxdb_port": 8086,
        "influxdb_username": None,
        "influxdb_password": None,
        "influxdb_ssl": False,
        "influxdb_verify_ssl": False,
    }
    # Temp comment for now
    # omotes-poc-test.hesi.energy
    # port 8086
    # user write-user
    # password nwn_write_test

    _ = run_optimization_problem(
        EndScenarioSizingHIGHS,
        esdl_run_info_path=runinfo_path,
        log_level=log_level,
        **kwargs,
    )


if __name__ == "__main__":
    main()



from mesido.pycml import Variable

from numpy import nan

from .electricity_base import ElectricityPort
from .._internal import BaseAsset
from .._internal.electricity_component import ElectricityComponent


class ElectricityDemand(ElectricityComponent, BaseAsset):
    """
    The electricity demand models consumption of electrical power. We set an equality constriant
    in to enforce the minimum voltage and the associated power at the demand. This allows us to
    overestimate the power losses in the rest of the network.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "electricity_demand"
        self.min_voltage = nan
        self.elec_power_nominal = nan
        self.id_mapping_carrier = -1

        self.add_variable(ElectricityPort, "ElectricityIn")
        self.add_variable(Variable, "Electricity_demand", min=0.0, nominal=self.elec_power_nominal)

        self.add_equation(
            ((self.ElectricityIn.Power - self.Electricity_demand) / self.elec_power_nominal)
        )



import logging

import esdl

from mesido.network_common import NetworkSettings

import numpy as np

from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)

logger = logging.getLogger("mesido")


class ESDLAdditionalVarsMixin(CollocatedIntegratedOptimizationProblem):
    __temperature_options = {}

    def read(self):
        super().read()

        for asset in [
            *self.energy_system_components.get("heat_source", []),
            *self.energy_system_components.get("ates", []),
            *self.energy_system_components.get("heat_buffer", []),
            *self.energy_system_components.get("heat_pump", []),
        ]:
            esdl_asset = self.esdl_assets[self.esdl_asset_name_to_id_map[asset]]
            for constraint in esdl_asset.attributes.get("constraint", []):
                if constraint.name == "setpointconstraint":
                    time_unit = constraint.range.profileQuantityAndUnit.perTimeUnit
                    if time_unit == esdl.TimeUnitEnum.HOUR:
                        time_hours = 1
                    elif time_unit == esdl.TimeUnitEnum.DAY:
                        time_hours = 24
                    elif time_unit == esdl.TimeUnitEnum.WEEK:
                        time_hours = 24 * 7
                    elif time_unit == esdl.TimeUnitEnum.MOTH:
                        time_hours = 24 * 7 * 30
                    elif time_unit == esdl.TimeUnitEnum.DAY:
                        time_hours = 365 * 24
                    else:
                        logger.error(
                            f"{asset} has a setpoint constaint specified with unknown"
                            f"per time unit"
                        )

                    asset_setpoints = {asset: (time_hours, constraint.range.maxValue)}

                    self._timed_setpoints.update(asset_setpoints)

    def pipe_classes(self, p):
        return self._override_pipe_classes.get(p, [])

    def esdl_heat_model_options(self):
        """Overwrites the fraction of the minimum tank volume"""
        options = super().esdl_heat_model_options()
        options["min_fraction_tank_volume"] = 0.0
        return options

    def temperature_carriers(self):
        return self.esdl_carriers_typed(type=[str(NetworkSettings.NETWORK_TYPE_HEAT).lower()])

    def electricity_carriers(self):
        return self.esdl_carriers_typed(
            type=[str(NetworkSettings.NETWORK_TYPE_ELECTRICITY).lower()]
        )

    def gas_carriers(self):
        return self.esdl_carriers_typed(
            type=[
                str(NetworkSettings.NETWORK_TYPE_GAS).lower(),
                str(NetworkSettings.NETWORK_TYPE_HYDROGEN).lower(),
            ]
        )

    def temperature_regimes(self, carrier):
        temperature_options = []
        temperature_step = 2.5

        try:
            temperature_options = self.__temperature_options[carrier]
        except KeyError:
            for asset in [
                *self.energy_system_components.get("heat_source", []),
                *self.energy_system_components.get("ates", []),
                *self.energy_system_components.get("heat_buffer", []),
                *self.energy_system_components.get("heat_pump", []),
                *self.energy_system_components.get("heat_exchanger", []),
                *self.energy_system_components.get("heat_demand", []),
            ]:
                esdl_asset = self.esdl_assets[self.esdl_asset_name_to_id_map[asset]]
                parameters = self.parameters(0)
                for i in range(len(esdl_asset.attributes["constraint"].items)):
                    constraint = esdl_asset.attributes["constraint"].items[i]
                    if (
                        constraint.name == "supply_temperature"
                        and carrier == parameters[f"{asset}.T_supply_id"]
                    ) or (
                        constraint.name == "return_temperature"
                        and carrier == parameters[f"{asset}.T_return_id"]
                    ):
                        try:
                            lb = self.__temperature_options[carrier][0]
                            ub = self.__temperature_options[carrier][-1]
                            lb = constraint.range.minValue if constraint.range.minValue > lb else lb
                            ub = constraint.range.maxValue if constraint.range.maxValue < ub else ub
                        except KeyError:
                            lb = constraint.range.minValue
                            ub = constraint.range.maxValue
                        n_options = round((ub - lb) / temperature_step)
                        temperature_options = np.linspace(lb, ub, n_options + 1)
                        if (
                            constraint.range.profileQuantityAndUnit.unit
                            != esdl.UnitEnum.DEGREES_CELSIUS
                        ):
                            RuntimeError(
                                f"{asset} has a temperature constraint with wrong unit "
                                f"{constraint.range.profileQuantityAndUnit.unit}, should "
                                f"always be in degrees celcius."
                            )
                        self.__temperature_options[carrier] = temperature_options

        return temperature_options



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestColdDemand(TestCase):
    def test_cold_demand(self):
        """
        This test is to check the basic physics for a network which includes cold demand. In this
        case we have a network with an air-water hp, a low temperature ates and both hot and cold
        demand. In this case the demands are matched and the low temperature ates is utilized.

        Checks:
        1. demand is matched
        2. energy conservation in the network
        3. heat to discharge (note cold line is colder than T_ground)

        """
        import models.wko.src.example as example
        from models.wko.src.example import HeatProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="LT_wko.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        demand_matching_test(heat_problem, heat_problem.extract_results())
        energy_conservation_test(heat_problem, heat_problem.extract_results())
        heat_to_discharge_test(heat_problem, heat_problem.extract_results())

    def test_airco(self):
        """
        This test is to check the basic physics for a network which includes an airco. In this
        case we have a network with an air-water hp, a low temperature ates and both hot and cold
        demand. In this case the demands are matched and the low temperature ates is utilized.

        Checks:
        1. demand is matched
        2. energy conservation in the network
        3. heat to discharge

        """
        import models.wko.src.example as example
        from models.wko.src.example import HeatProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="airco.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        demand_matching_test(heat_problem, heat_problem.extract_results())
        energy_conservation_test(heat_problem, heat_problem.extract_results())
        heat_to_discharge_test(heat_problem, heat_problem.extract_results())


if __name__ == "__main__":
    a = 1



import subprocess
import sys
from pathlib import Path

import pytest

examples = list((Path(__file__).resolve().parent.parent / "examples").glob("**/src/example.py"))


@pytest.mark.parametrize("example", examples, ids=[x.parent.parent.name for x in examples])
def test_examples(example):
    """
    Check if the examples provided in the repo still are succeeding the optimization.

    """
    example_name = example.parent.parent.name
    env = sys.executable
    try:
        subprocess.check_output([env, str(example)])
    except Exception:
        pytest.fail(f"Running example '{example_name}' failed")



from mesido.pycml import Connector, Variable

from .._internal.gas_component import GasComponent


class GasPort(GasComponent, Connector):
    """
    The gas port is used to model the variables at a port where two assets are connected. For the
    gas network we model flow (Q [m3/s]) and head (H [m]). The Q_shadow variable is only used for
    correctly connecting ports of assets later on.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)
        # TODO: think of more elegant approach for Q_shadow, currently required to ensure that
        #  every port has a unique variable to make the correct port mapping
        self.add_variable(Variable, "Q")  # [m3/s]
        self.add_variable(Variable, "Q_shadow")
        self.add_variable(Variable, "mass_flow")  # [g/s]
        self.add_variable(Variable, "H")  # [m]
        self.add_variable(Variable, "Hydraulic_power")  # [W]


class GasTwoPort(GasComponent):
    """
    For gas components that transport flow we have a two port component.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(GasPort, "GasIn")
        self.add_variable(GasPort, "GasOut")



import sys
from pathlib import Path
from typing import Dict
from unittest import TestCase

from mesido._heat_loss_u_values_pipe import pipe_heat_loss
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.pipe_class import PipeClass
from mesido.techno_economic_mixin import TechnoEconomicMixin
from mesido.util import run_esdl_mesido_optimization

import numpy as np
import numpy.testing

MIP_TOLERANCE = 1e-8


class TestTopoConstraintsOnPipeDiameterSizingExample(TestCase):
    """
    Tests the topo variables and constraints of heat_mixin on the Pipe Diameter Sizing example.
    """

    problem: TechnoEconomicMixin
    results: Dict[str, np.ndarray]

    @classmethod
    def setUpClass(cls) -> None:
        root_folder = str(Path(__file__).resolve().parent.parent)
        sys.path.insert(1, root_folder)

        import examples.pipe_diameter_sizing.src.example  # noqa: E402, I100
        from examples.pipe_diameter_sizing.src.example import (
            PipeDiameterSizingProblem,
        )  # noqa: E402, I100

        base_folder = (
            Path(examples.pipe_diameter_sizing.src.example.__file__).resolve().parent.parent
        )

        del root_folder
        sys.path.pop(1)

        cls.problem = run_esdl_mesido_optimization(
            PipeDiameterSizingProblem,
            base_folder=base_folder,
            esdl_file_name="2a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )
        cls.results = cls.problem.extract_results()

    def test_pipe_class_var(self):
        """
        This test is to check whether all variables associated to pipe class optimization are set
        as expected.

        Tests the variables stored in:
        - pipe_topo_pipe_class_var check that only one is selected
        - pipe_topo_max_discharge_var is that of the selected pipe class
        - pipe_topo_cost_var is that of the selected pipe class
        - pipe_diameter_var is that of the selected pipe class
        - pipe_heat_loss_var is that of the selected pipe class
        """
        for p in self.problem.energy_system_components.get("heat_pipe", []):
            # If there is nothing to choose for the optimizer, no pipe class binaries are made
            if self.problem.pipe_classes(p) is None or len(self.problem.pipe_classes(p)) == 1:
                continue

            class_vars = self.get_pipe_class_vars(p)
            chosen_pc = self.get_chosen_pipe_class(p, class_vars)

            for var_name, value in class_vars.items():
                self.assertTrue(var_name in self.results, msg=f"{var_name} not in results")
                self.assertTrue(
                    abs(value - 0.0) < MIP_TOLERANCE or abs(value - 1.0) < MIP_TOLERANCE,
                    msg=f"Binary {var_name} isn't either 0.0 or 1.0, it is {value}",
                )
            np.testing.assert_almost_equal(
                1.0,
                np.sum(val for val in class_vars.values()),
                err_msg=f"Not exactly 1 pipe class selected for {p}",
            )

            np.testing.assert_array_almost_equal(
                chosen_pc.inner_diameter,
                self.results[f"{p}__hn_diameter"],
                err_msg=f"{p} inner diameter doesn't match expected",
            )
            np.testing.assert_array_almost_equal(
                chosen_pc.investment_costs,
                self.results[f"{p}__hn_cost"],
                err_msg=f"{p} investment costs doesn't match expected",
            )
            np.testing.assert_array_almost_equal(
                chosen_pc.maximum_discharge,
                self.results[f"{p}__hn_max_discharge"],
                err_msg=f"{p} maximum discharge doesn't match expected",
            )

            expected_heat_losses = self.get_heat_losses(p, chosen_pc)
            np.testing.assert_almost_equal(
                self.results[f"{p}__hn_heat_loss"], expected_heat_losses, 5
            )

    def test_pipe_class_ordering_vars(self):
        """
        This test is to check if the pipe class ordering variables are set as expected. The pipe
        class ordering variables are there to help the optimizer in seeing the relation between
        mulitple integer variables.

        Tests the variables stored in:
        - pipe_topo_global_pipe_clas_count_var
        - pipe_topo_pipe_class_discharge_ordering_var
        - pipe_topo_pipe_class_cost_ordering_var
        - pipe_topo_pipe_class_heat_loss_ordering_var
        """
        pc_sums = {pc.name: 0 for pc in self.problem.get_unique_pipe_classes()}
        total_pipes_to_optimize = 0
        for p in self.problem.energy_system_components.get("heat_pipe", []):
            # If there is nothing to choose for the optimizer, no pipe class binaries are made,
            if self.problem.pipe_classes(p) is None or len(self.problem.pipe_classes(p)) == 1:
                continue
            pipe_class_vars = self.get_pipe_class_vars(p)
            chosen_pc = self.get_chosen_pipe_class(p, pipe_class_vars)
            pc_sums[chosen_pc.name] += 1
            total_pipes_to_optimize += 1
            for pc in self.problem.pipe_classes(p):
                if p in self.problem.cold_pipes:
                    base_name = f"{self.problem.cold_to_hot_pipe(p)}__hn_pipe_class_{pc.name}"
                else:
                    base_name = f"{p}__hn_pipe_class_{pc.name}"
                cost_ordering_var_name = base_name + "_cost_ordering"
                cost_ordering_var = self.results[cost_ordering_var_name]
                if pc.investment_costs < chosen_pc.investment_costs:
                    np.testing.assert_almost_equal(
                        cost_ordering_var,
                        0.0,
                        err_msg=f"expected the cost order var for {p} and {pc=} to be 0.0, since"
                        f"{chosen_pc=} with higher investment costs",
                    )
                elif pc.investment_costs > chosen_pc.investment_costs:
                    np.testing.assert_almost_equal(
                        cost_ordering_var,
                        1.0,
                        err_msg=f"Expected the cost order var for {p} and {pc=} to be 1.0, since"
                        f"{chosen_pc=} with lower investment costs",
                    )

                discharge_ordering_var_name = base_name + "_discharge_ordering"
                discharge_ordering_var = self.results[discharge_ordering_var_name]
                if pc.maximum_discharge < chosen_pc.maximum_discharge:
                    np.testing.assert_almost_equal(
                        discharge_ordering_var,
                        0.0,
                        err_msg=f"expected the discharge order var for {p} and {pc=} to be 0.0, "
                        f"since {chosen_pc=} with higher max discharge",
                    )
                elif pc.maximum_discharge > chosen_pc.maximum_discharge:
                    np.testing.assert_almost_equal(
                        discharge_ordering_var,
                        1.0,
                        err_msg=f"Expected the discharge order var for {p} and {pc=} to be 1.0, "
                        f"since {chosen_pc=} with lower max discharge",
                    )

                heat_loss_ordering_var_name = base_name + "_heat_loss_ordering"
                heat_loss_ordering_var = self.results[heat_loss_ordering_var_name]
                pc_heat_loss = self.get_heat_losses(p, pc)
                chosen_pc_heat_loss = self.get_heat_losses(p, chosen_pc)
                if pc_heat_loss < chosen_pc_heat_loss:
                    np.testing.assert_almost_equal(
                        heat_loss_ordering_var,
                        0.0,
                        err_msg=f"expected the heat loss order var for {p} and {pc=} to be 0.0, "
                        f"since {chosen_pc=} with higher heat losses",
                    )
                elif pc_heat_loss > chosen_pc_heat_loss:
                    np.testing.assert_almost_equal(
                        discharge_ordering_var,
                        1.0,
                        err_msg=f"Expected the heat loss order var for {p} and {pc=} to be 1.0, "
                        f"since {chosen_pc=} with lower heat losses",
                    )

        for pc_name, total_count in pc_sums.items():
            count_var_value = self.results[f"{pc_name}__global_pipe_class_count"]
            np.testing.assert_almost_equal(
                count_var_value,
                total_count,
                err_msg=f"Pipe count for {pc_name} doesn't match the expected {total_count}",
            )
        total_pipe_count = sum(pc_sums.values())
        np.testing.assert_equal(
            total_pipe_count,
            total_pipes_to_optimize,
            err_msg=f"Found {total_pipe_count} total selected pipe classes, "
            f"but {total_pipes_to_optimize=}",
        )

    def get_pipe_class_vars(self, pipe: str) -> Dict[str, np.ndarray]:
        """
        This function returns the pipe class results.

        Parameters
        ----------
        pipe : str with pipe name

        Returns
        -------
        Dict with variable name as key and result as value.
        """
        given_pipe_classes = self.problem.pipe_classes(pipe)
        if pipe in self.problem.cold_pipes:
            expected_class_vars = [
                f"{self.problem.cold_to_hot_pipe(pipe)}__hn_pipe_class_{pc.name}"
                for pc in given_pipe_classes
            ]
        else:
            expected_class_vars = [f"{pipe}__hn_pipe_class_{pc.name}" for pc in given_pipe_classes]
        class_vars = {
            var_name: value
            for var_name, value in self.results.items()
            if var_name in expected_class_vars
        }
        return class_vars

    def get_chosen_pipe_class(self, pipe: str, pipe_class_vars: Dict[str, np.ndarray]) -> PipeClass:
        """
        This function retrieves the selected pipe class optimization result for a pipe.

        Parameters
        ----------
        pipe : string with pipe name
        pipe_class_vars : dict from get_pipe_class_vars() method

        Returns
        -------
        The selected pipe class
        """
        chosen_var = None
        given_pipe_classes = self.problem.pipe_classes(pipe)
        for var_name, value in pipe_class_vars.items():
            if abs(value - 1.0) < MIP_TOLERANCE:
                chosen_var = var_name
        self.assertIsNotNone(chosen_var, msg=f"No pipe class selected for {pipe}")
        class_name = chosen_var.split("_")[-1]
        chosen_pc = [pc for pc in given_pipe_classes if pc.name == class_name]
        self.assertEqual(len(chosen_pc), 1, msg=f"Found multiple chosen pipe classes for {pipe}")
        return chosen_pc[0]

    def get_heat_losses(self, pipe: str, pipe_class: PipeClass):
        """
        This function computes the expected heat loss for a pipe class.

        Parameters
        ----------
        pipe : string with pipe name
        pipe_class : the selected pipe class optimized result for that pie

        Returns
        -------
        Pipe heat loss value.
        """
        return pipe_heat_loss(
            self.problem,
            options=self.problem.energy_system_options(),
            parameters=self.problem.parameters(0),
            p=pipe,
            u_values=pipe_class.u_values,
        )



from .model_base import (
    Component,
    Connector,
    ConstantInput,
    ControlInput,
    FlattenedModel,
    Model,
    SymbolicParameter,
    Variable,
)

__all__ = [
    "Component",
    "Connector",
    "ConstantInput",
    "ControlInput",
    "FlattenedModel",
    "Model",
    "SymbolicParameter",
    "Variable",
]



from .. import ElectricitySource


class WindPark(ElectricitySource):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_subtype = "wind_park"



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np


class TestElectricityTopo(TestCase):
    def test_electricity_network_topology(self):
        """
        This test checks the functioning of topology optimization of electricity cables. It uses
        a symmetrical network where the left side should stay in as those cables are shorter.

        Checks:
        1. Demand is matched
        2. That intended cables are removed

        """
        import models.electricity_cable_topology.src.example as example
        from models.electricity_cable_topology.src.example import HeatProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="enettopology.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        results = heat_problem.extract_results()

        for demand in heat_problem.energy_system_components.get("electricity_demand", []):
            target = heat_problem.get_timeseries(f"{demand}.target_electricity_demand").values
            np.testing.assert_allclose(target, results[f"{demand}.Electricity_demand"])

        removed_cables = [
            "ElectricityCable_6dce",
            "ElectricityCable_6aeb",
            "ElectricityCable_fbfd",
            "ElectricityCable_ef84",
        ]

        for cable in removed_cables:
            np.testing.assert_allclose(results[f"{cable}__investment_cost"], 0.0)
            np.testing.assert_allclose(results[f"{cable}__en_max_current"], 0.0)



from mesido.pycml import Variable
from mesido.pycml.component_library.milp._internal import BaseAsset
from mesido.pycml.component_library.milp.heat.heat_four_port import HeatFourPort

from numpy import nan


class HeatPump(HeatFourPort, BaseAsset):
    """
    The heatpump component is used to model a water-water heatpump.
    A constant COP is used to model the electricity use of the heatpump. A power cap is set
    on the primary side to model physical constraints on the amount of heat transfer.

    The heat to discharge constraints are set in the HeatPhysicsMixin. The primary side is modelled
    as a demand, meaning it consumes energy from the primary network and gives it to the secondary
    side, where the secondary side acts like a source to the secondary network. This also means
    that heat can only flow from primary to secondary.
    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(),
                modifiers,
            ),
        )

        self.component_type = "heat_pump"
        self.efficiency = nan
        self.COP = nan  # TODO: maybe set this to a standard value if not set in esdl.
        self.minimum_pressure_drop = 1.0e5  # 1 bar of pressure drop
        self.nominal = (
            self.Secondary.Q_nominal * self.Secondary.rho * self.Secondary.cp * self.Secondary.dT
        )
        self.pump_efficiency = 0.5
        self.elec_power_nominal = self.nominal / self.COP

        # Assumption: heat in/out and added is nonnegative

        self.add_variable(Variable, "Primary_heat", min=0.0)
        self.add_variable(Variable, "Secondary_heat", min=0.0)
        self.add_variable(Variable, "Heat_flow", nominal=self.nominal)
        self.add_variable(Variable, "Power_elec", min=0.0)
        self.add_variable(Variable, "dH_prim", max=0.0)
        self.add_variable(Variable, "dH_sec", min=0.0)

        # Hydraulically decoupled so Heads remain the same
        # #TODO: can't these two equations be moved to the non_storagecomponent?
        self.add_equation(self.dH_prim - (self.Primary.HeatOut.H - self.Primary.HeatIn.H))
        self.add_equation(
            (
                self.minimum_pressure_drop * self.Primary.Q
                - (self.Primary.HeatIn.Hydraulic_power - self.Primary.HeatOut.Hydraulic_power)
            )
            / (self.Primary.Q_nominal * self.Primary.nominal_pressure)
        )
        self.add_equation(self.dH_sec - (self.Secondary.HeatOut.H - self.Secondary.HeatIn.H))
        self.add_equation(
            (
                self.Pump_power
                - (self.Secondary.HeatOut.Hydraulic_power - self.Secondary.HeatIn.Hydraulic_power)
            )
            / (self.Secondary.Q_nominal * self.Secondary.nominal_pressure)
        )

        self.add_equation(
            ((self.Primary_heat + self.Power_elec - self.Secondary_heat) / self.nominal)
        )

        self.add_equation(
            (
                (self.Primary_heat - (self.Primary.HeatIn.Heat - self.Primary.HeatOut.Heat))
                / self.nominal
            )
        )
        self.add_equation(
            (
                (self.Secondary_heat - (self.Secondary.HeatOut.Heat - self.Secondary.HeatIn.Heat))
                / self.nominal
            )
        )
        self.add_equation((self.Heat_flow - self.Secondary_heat) / self.nominal)



import argparse
import logging
import re
import sys
import time
from pathlib import Path

import esdl

from mesido import __version__


MULTI_ENUM_NAME_TO_FACTOR = {
    esdl.MultiplierEnum.ATTO: 1e-18,
    esdl.MultiplierEnum.FEMTO: 1e-15,
    esdl.MultiplierEnum.PICO: 1e-12,
    esdl.MultiplierEnum.NANO: 1e-9,
    esdl.MultiplierEnum.MICRO: 1e-6,
    esdl.MultiplierEnum.MILLI: 1e-3,
    esdl.MultiplierEnum.CENTI: 1e-2,
    esdl.MultiplierEnum.DECI: 1e-1,
    esdl.MultiplierEnum.NONE: 1e0,
    esdl.MultiplierEnum.DEKA: 1e1,
    esdl.MultiplierEnum.HECTO: 1e2,
    esdl.MultiplierEnum.KILO: 1e3,
    esdl.MultiplierEnum.MEGA: 1e6,
    esdl.MultiplierEnum.GIGA: 1e9,
    esdl.MultiplierEnum.TERA: 1e12,
    esdl.MultiplierEnum.TERRA: 1e12,
    esdl.MultiplierEnum.PETA: 1e15,
    esdl.MultiplierEnum.EXA: 1e18,
}


def _sort_numbered(list_):
    """
    If there are any integers in the string, we want to parse it as a
    number. That way, "Pipe2" comes before "Pipe10".
    """
    return sorted(
        list_, key=lambda x: tuple(int(y) if y.isdigit() else y for y in re.split(r"(\d+)$", x))
    )


def main_decorator(func):
    def main(runinfo_path=None, log_level=None, run_remote=None):
        logger = logging.getLogger("WarmingUP-MPC")
        logger.setLevel(logging.INFO)

        def handle_exception(exc_type, exc_value, exc_traceback):
            if issubclass(exc_type, KeyboardInterrupt):
                sys.__excepthook__(exc_type, exc_value, exc_traceback)
                return

            import traceback

            fmt_tb = "".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
            logging.exception(f"Uncaught exception: {fmt_tb}")

        sys.excepthook = handle_exception

        start_time = time.time()

        if func.__module__ == "__main__":
            parser = argparse.ArgumentParser(description="Run ESDL model")
            parser.add_argument(
                "-l",
                "--log",
                default="WARN",
                dest="log_level",
                choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
                help="Set the logging level (default: %(default)s)",
            )

            remote_parser = parser.add_mutually_exclusive_group(required=False)
            remote_parser.add_argument(
                "--remote", dest="remote", action="store_true", help="Run the workflow remotely"
            )
            remote_parser.add_argument(
                "--local",
                dest="remote",
                action="store_false",
                help="Run the workflow locally (default)",
            )
            parser.set_defaults(remote=False)

            arguments = parser.parse_args()

            if runinfo_path is None:
                runinfo_path = Path(arguments.runinfo_path)
            if log_level is None:
                log_level = logging.getLevelName(arguments.log_level)
            if run_remote is None:
                run_remote = arguments.remote

        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s",
            level=log_level,
            handlers=[logging.StreamHandler()],
        )

        logger.info(f"Using WarmingUP-MPC {__version__}.")

        if run_remote:
            import inspect
            from warmingup_mpc.amqp_client import AMQPClient

            workflow_name = Path(inspect.getfile(func)).stem

            mpc_rpc = AMQPClient()

            # Effectively calling "func" remotely
            mpc_rpc.run_runinfo_workflow(workflow_name, runinfo_path, log_level)
        else:
            func(runinfo_path, log_level)

        print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))

    return main


def get_cost_value_and_unit(cost_info: esdl.SingleValue):
    cost_value = cost_info.value
    unit_info = cost_info.profileQuantityAndUnit
    unit = unit_info.unit
    per_time_uni = unit_info.perTimeUnit
    per_unit = unit_info.perUnit
    multiplier = unit_info.multiplier
    per_multiplier = unit_info.perMultiplier

    cost_value *= MULTI_ENUM_NAME_TO_FACTOR[multiplier]
    cost_value /= MULTI_ENUM_NAME_TO_FACTOR[per_multiplier]

    return cost_value, unit, per_unit, per_time_uni



from mesido.esdl.esdl_parser import BaseESDLParser
from mesido.esdl.profile_parser import BaseProfileReader, InfluxDBProfileReader

from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.util import run_optimization_problem


def run_esdl_mesido_optimization(
    problem: CollocatedIntegratedOptimizationProblem,
    esdl_parser: BaseESDLParser,
    esdl_file_name: str = "",
    base_folder: str = "",
    esdl_string: str = "",
    profile_reader: BaseProfileReader = InfluxDBProfileReader,
    input_timeseries_file: str = "",
    *args,
    **kwargs,
):
    """
    This function is used to execute an optimization on a mesido defined problem. Compared to the
    standard rtc-tools run_optimization_problem() method extra checks can be included here.

    params:
    Problem: The problem defined in the mesido
    esdl_parser: How the ESDL will be provided to the problem, e.g. file or base64 string.
    base_folder: The base folder from where the input, model, output folder are defined
    esdl_file_name: The string of the esdl file in case the ESDLFileParser is used and placed in
    the model folder.
    esdl_string: the base64 string in case the ESDLStringParser is selected
    profile_reader: The way the time-series profiles are read
    input_timeseries_file: The file from which to read the time-series profiles in case the
    FromFileReader is selected.

    Returns:
    The solved full problem object with the solution in it.

    """

    solution = run_optimization_problem(
        problem,
        *args,
        base_folder=base_folder,
        esdl_file_name=esdl_file_name,
        esdl_string=esdl_string,
        esdl_parser=esdl_parser,
        profile_reader=profile_reader,
        input_timeseries_file=input_timeseries_file,
        **kwargs,
    )

    feasibility = solution.solver_stats["return_status"]

    assert feasibility.lower() == "optimal"

    return solution



import logging

from mesido.asset_sizing_mixin import AssetSizingMixin
from mesido.base_component_type_mixin import BaseComponentTypeMixin
from mesido.financial_mixin import FinancialMixin
from mesido.physics_mixin import PhysicsMixin

from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)

logger = logging.getLogger("mesido")


class TechnoEconomicMixin(
    FinancialMixin,
    AssetSizingMixin,
    PhysicsMixin,
    BaseComponentTypeMixin,
    CollocatedIntegratedOptimizationProblem,
):
    """
    This class combines the different Mixin classes needed to do a full techno-economic
    optimization. This class is created so that the different Mixin are fully modular and users
    of the code base are able to replace a specific Mixin with their own implementation without
    needing a full refactoring of the code.

    """

    def __init__(self, *args, **kwargs):
        """
        In this __init__ we prepare the dicts for the variables added by the HeatMixin class
        """

        super().__init__(*args, **kwargs)

    def pre(self):
        """
        In this pre method we fill the dicts initiated in the __init__. This means that we create
        the Casadi variables and determine the bounds, nominals and create maps for easier
        retrieving of the variables.
        """
        super().pre()

    def get_max_size_var(self, asset_name, ensemble_member):
        return self.extra_variable(self._asset_max_size_map[asset_name], ensemble_member)

    def get_aggregation_count_var(self, asset_name, ensemble_member):
        return self.extra_variable(
            self._asset_aggregation_count_var_map[asset_name], ensemble_member
        )

    def get_aggregation_count_max(self, asset_name):
        return self.bounds()[self._asset_aggregation_count_var_map[asset_name]][1]

    def get_pipe_class_map(self):
        return self._heat_pipe_topo_pipe_class_map

    def get_gas_pipe_class_map(self):
        return self._gas_pipe_topo_pipe_class_map

    def get_electricity_cable_class_map(self):
        return self._electricity_cable_topo_cable_class_map

    def get_pipe_investment_cost_coefficient(self, asset_name, ensemble_member):
        return self.extra_variable(self._heat_pipe_topo_cost_map[asset_name], ensemble_member)

    def get_electricity_carriers(self):
        return self.electricity_carriers()

    def get_gas_carriers(self):
        return self.gas_carriers()

    def get_heat_carriers(self):
        return self.temperature_carriers()

    def get_gas_pipe_investment_cost_coefficient(self, asset_name, ensemble_member):
        return self.extra_variable(self._gas_pipe_topo_cost_map[asset_name], ensemble_member)

    def get_electricity_cable_investment_cost_coefficient(self, asset_name, ensemble_member):
        return self.extra_variable(
            self._electricity_cable_topo_cost_map[asset_name], ensemble_member
        )

    def energy_system_options(self):
        r"""
        Returns a dictionary of milp network specific options.
        """

        options = PhysicsMixin.energy_system_options(self)
        options.update(FinancialMixin.energy_system_options(self))
        # problem with abstractmethod
        options["include_asset_is_realized"] = False

        return options

    @property
    def extra_variables(self):
        """
        In this function we add all the variables defined in the HeatMixin to the optimization
        problem. Note that these are only the normal variables not path variables.
        """
        variables = super().extra_variables.copy()

        return variables

    @property
    def path_variables(self):
        """
        In this function we add all the path variables defined in the HeatMixin to the
        optimization problem. Note that path_variables are variables that are created for each
        time-step.
        """
        variables = super().path_variables.copy()

        return variables

    def variable_is_discrete(self, variable):
        """
        All variables that only can take integer values should be added to this function.
        """

        return super().variable_is_discrete(variable)

    def variable_nominal(self, variable):
        """
        In this function we add all the nominals for the variables defined/added in the HeatMixin.
        """

        return super().variable_nominal(variable)

    def bounds(self):
        """
        In this function we add the bounds to the problem for all the variables defined/added in
        the HeatMixin.
        """
        bounds = super().bounds()

        return bounds

    def path_constraints(self, ensemble_member):
        """
        Here we add all the path constraints to the optimization problem. Please note that the
        path constraints are the constraints that are applied to each time-step in the problem.
        """

        constraints = super().path_constraints(ensemble_member)

        return constraints

    def constraints(self, ensemble_member):
        """
        This function adds the normal constraints to the problem. Unlike the path constraints these
        are not applied to every time-step in the problem. Meaning that these constraints either
        consider global variables that are independent of time-step or that the relevant time-steps
        are indexed within the constraint formulation.
        """
        constraints = super().constraints(ensemble_member)

        return constraints

    def goal_programming_options(self):
        """
        Here we set the goal programming configuration. We use soft constraints for consecutive
        goals.
        """
        options = super().goal_programming_options()
        options["keep_soft_constraints"] = True
        return options

    def solver_options(self):
        """
        Here we define the solver options. By default we use the open-source solver highs and casadi
        solver qpsol.
        """
        options = super().solver_options()
        options["casadi_solver"] = "qpsol"
        options["solver"] = "highs"
        return options

    def compiler_options(self):
        """
        In this function we set the compiler configuration.
        """
        options = super().compiler_options()
        options["resolve_parameter_values"] = True
        return options



import logging
from abc import abstractmethod
from enum import IntEnum
from typing import List, Optional, Tuple, Type, Union

import casadi as ca

import mesido._darcy_weisbach as darcy_weisbach
from mesido.constants import GRAVITATIONAL_CONSTANT
from mesido.network_common import NetworkSettings
from mesido.pipe_class import PipeClass

import numpy as np

from rtctools.optimization.goal_programming_mixin_base import Goal
from rtctools.optimization.optimization_problem import BT


logger = logging.getLogger("mesido")


def _state_vector_scaled(optimization_problem, variable, ensemble_member):
    """
    This functions returns the casadi symbols scaled with their nominal for the entire time
    horizon.
    """
    canonical, sign = optimization_problem.alias_relation.canonical_signed(variable)
    return (
        optimization_problem.state_vector(canonical, ensemble_member)
        * optimization_problem.variable_nominal(canonical)
        * sign
    )


class HeadLossOption(IntEnum):
    r"""
    Enumeration for the possible options to take head loss in pipes into account.
    Also see :py:meth:`._HeadLossMixin.energy_system_options` for related options.

    .. note::
        Not all options are supported by :py:class:`.HeatMixin`, due to the focus
        on MILP formulations.

    NO_HEADLOSS
       The NO_HEADLOSS option assumes that there is no headloss in the pipelines.
       There are no constraints added relating the discharge to the head.

    CQ2_INEQUALITY
        As the name implies, this adds a quadratic inquality constraint between
        the head and the discharge in a pipe:

         .. math::

           dH \ge C \cdot Q^2

        This expression of the headloss requires a system-specific estimation of
        the constant C.

        As dH is always positive, a boolean is needed when flow directions are not
        fixed in a mixed-integer formulation to determine if

         .. math::

           dH = H_{up} - H_{down}

        or (when the :math:`Q < 0`)

         .. math::

           dH = H_{down} - H_{up}

    LINEARIZED_N_LINES_WEAK_INEQUALITY
        Just like ``CQ2_INEQUALITY``, this option adds inequality constraints:

        .. math::
           \Delta H \ge \vec{a} \cdot Q + \vec{b}

        with :math:`\vec{a}` and :math:`\vec{b}` the linearization coefficients.

        This approach can more easily be explain with a plot, showing the Darcy-Weisbach
        head loss, and the linear lines approximating it. Note that the number of
        supporting lines is an option that can be set by the user by overriding
        :py:meth:`._HeadLossMixin.energy_system_options`. Also note that, just like
        ``CQ2_INEQUALITY``, a boolean is needed when flow directions are not fixed.

           .. image:: /images/DWlinearization.PNG

    LINEARIZED_ONE_LINE_EQUALITY
        This option uses a linear head loss formulation.
        A single constraint of the type

         .. math::

           H_{up} - H_{down} = dH = C \cdot Q

        is added.
        Note that no boolean are required to support the case where flow directions
        are not fixed yet, at the cost of reduced fidelity in the head-loss relationship.

        The exact velocity to use to linearize can be set by overriding
        :py:meth:`._HeadLossMixin.energy_system_options`.

    CQ2_EQUALITY
        This option adds **equality** constraints of the type:

         .. math::

           dH = C \cdot Q^2

        This equation is non-convex, and can therefore lead to convergence issues.
    """

    NO_HEADLOSS = 1
    CQ2_INEQUALITY = 2
    LINEARIZED_N_LINES_WEAK_INEQUALITY = 3
    LINEARIZED_ONE_LINE_EQUALITY = 4
    CQ2_EQUALITY = 5
    LINEARIZED_N_LINES_EQUALITY = 6


class _MinimizeHeadLosses(Goal):
    order = 1

    priority = 2**31 - 2

    def __init__(self, optimization_problem, input_network_settings, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.optimization_problem = optimization_problem
        self.network_settings = input_network_settings
        self.function_nominal = len(optimization_problem.times())

    def function(self, optimization_problem, ensemble_member):
        """
        This function returns the summed head loss of all pipes and pumps.
        """
        sum_ = 0.0

        parameters = optimization_problem.parameters(ensemble_member)

        pumps = optimization_problem.energy_system_components.get("pump", [])
        sources = optimization_problem.energy_system_components.get("heat_source", [])

        for p in pumps:
            sum_ += optimization_problem.state(f"{p}.dH")

        # If sources have an accompanying pump, we prefer the produced head to
        # be shifted to that pump. We therefore penalize the head of the
        # sources twice as much.
        for s in sources:
            sum_ += 2 * optimization_problem.state(f"{s}.dH")

        assert self.network_settings["head_loss_option"] != HeadLossOption.NO_HEADLOSS

        if self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_HEAT:
            for p in optimization_problem.energy_system_components.get("heat_pipe", []):
                if (
                    not parameters[f"{p}.has_control_valve"]
                    and not parameters[f"{p}.length"] == 0.0
                ):
                    sym_name = optimization_problem._hn_pipe_to_head_loss_map[p]
                    sum_ += optimization_problem.state(sym_name)
        elif self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_GAS:
            for p in optimization_problem.energy_system_components.get("gas_pipe", []):
                if not parameters[f"{p}.length"] == 0.0:
                    sym_name = optimization_problem._gn_pipe_to_head_loss_map[p]
                    sum_ += optimization_problem.state(sym_name)

        return sum_


class _MinimizeHydraulicPower(Goal):
    order = 1

    priority = 2**31 - 1

    def __init__(
        self,
        optimization_problem,
        input_network_settings,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.optimization_problem = optimization_problem
        self.network_settings = input_network_settings

    def function(self, optimization_problem, ensemble_member):
        """
        This function returns the summed hydraulic power of all pipes
        """
        sum_ = 0.0

        parameters = optimization_problem.parameters(ensemble_member)

        assert self.network_settings["head_loss_option"] != HeadLossOption.NO_HEADLOSS

        for pipe in optimization_problem.energy_system_components.get("heat_pipe", []):
            if (
                not parameters[f"{pipe}.has_control_valve"]
                and not parameters[f"{pipe}.length"] == 0.0
            ):
                sum_ += optimization_problem.state(f"{pipe}.Hydraulic_power")

        return sum_


class HeadLossClass:
    """
    For handling of discharge - head (loss) relationship to the model.
    """

    def __init__(self, input_network_settings, *args, **kwargs):

        self.__pipe_head_bounds = {}

        self.__pipe_head_loss_var = {}
        self.__pipe_head_loss_bounds = {}
        self.__pipe_head_loss_nominals = {}
        self.__pipe_head_loss_zero_bounds = {}
        self._hn_pipe_to_head_loss_map = {}

        # Kvr
        # Boolean variables for the linear line segment options per pipe.
        self.__pipe_linear_line_segment_var = {}  # value 0/1: line segment - not active/active
        self.__pipe_linear_line_segment_var_bounds = {}
        self._pipe_linear_line_segment_map = {}

        self.__priority = None

        self.network_settings = input_network_settings

    def pre(self):
        """
        Some checks to avoid that different pipes have different head_loss options in case one
        has the No_HeadLoss option.
        """
        super().pre()

        self.__initialize_nominals_and_bounds()

        parameters = self.parameters(0)

        # It is not allowed to mix NO_HEADLOSS with other head loss options as
        # that just leads to weird and undefined behavior.
        head_loss_values = {
            self.network_settings["head_loss_option"],
        }

        pipe_type = "heat_pipe"
        if len(self.energy_system_components.get("heat_pipe", [])) == 0:
            pipe_type = "gas_pipe"

        for p in self.energy_system_components.get(pipe_type, []):
            head_loss_values.add(
                self._hn_get_pipe_head_loss_option(p, self.network_settings, parameters)
            )

        if HeadLossOption.NO_HEADLOSS in head_loss_values and len(head_loss_values) > 1:
            raise Exception(
                "Mixing .NO_HEADLOSS with other head loss options is not allowed. "
                "Either all pipes should have .NO_HEADLOSS set, or none. "
                "The global value returned by energy_system_options() also need to match."
            )

    def head_loss_network_options(self):
        r"""
        Returns a dictionary of milp network specific options.

        +--------------------------------+-----------+-----------------------------------+
        | Option                         | Type      | Default value                     |
        +================================+===========+===================================+
        | ``minimum_pressure_far_point`` | ``float`` | ``1.0`` bar                       |
        +--------------------------------+-----------+-----------------------------------+
        | ``wall_roughness``             | ``float`` | ``0.0002`` m                      |
        +--------------------------------+-----------+-----------------------------------+
        | ``estimated_velocity``         | ``float`` | ``1.0`` m/s (CQ2_* &              |
        |                                |           |LINEARIZED_ONE_LINE_EQUALITY)      |
        +--------------------------------+-----------+-----------------------------------+

        The ``minimum_pressure_far_point`` gives the minimum pressure
        requirement at any demand node, which means that the pressure at the
        furthest point is also satisfied without inspecting the topology.

        The ``wall_roughness`` of the pipes plays a role in determining the
        resistance of the pipes.
        """

        options = {}

        options["minimum_pressure_far_point"] = 1.0
        options["wall_roughness"] = 2e-4
        options["estimated_velocity"] = 1.0

        return options

    @abstractmethod
    def _hn_get_pipe_head_loss_option(
        self, pipe, network_settings, parameters, **kwargs
    ) -> HeadLossOption:
        """
        The global user head loss option is not necessarily the same as the
        head loss option for a specific pipe. For example, when a control
        valve is present, a .LINEARIZED_ONE_LINE_EQUALITY global head loss option could mean a
        .CQ2_INEQUALITY formulation should be used instead.

        See also the explanation of `head_loss_option` (and its values) in
        :py:meth:`.energy_system_options`.
        """
        raise NotImplementedError

    @abstractmethod
    def _hn_pipe_head_loss_constraints(self, ensemble_member) -> List[Tuple[ca.MX, float, float]]:
        """
        This method should be implemented to relate the three variables:

        - discharge: e.g. `pipe.Q`
        - head difference: e.g. `pipe.dH`
        - head loss of the pipe (note: proxy symbol that is >= abs(actual head loss))

        The internal variable name/symbol for the head loss can be retried via
        `self._hn_pipe_to_head_loss_map[pipe]`. The method that should/can be called to
        relate these three variables is :py:meth:`._hn_pipe_head_loss`.
        """
        raise NotImplementedError

    @property
    def _hn_minimization_goal_class(self) -> Type[Goal]:
        """
        This function returns the Minimize Head loss goal
        """
        return _MinimizeHeadLosses

    @property
    def _hpwr_minimization_goal_class(self) -> Type[Goal]:
        """
        This function returns the minimize hydraulic power goal
        """
        return _MinimizeHydraulicPower

    def initialize_variables_nominals_and_bounds(
        self, optimization_problem, commodity_type, pipe_name, network_settings
    ):
        """
        This function computes and sets the bounds and nominals for the head loss of all the pipes
        as well as the minimum and maximum pipe pressure.
        """
        options = optimization_problem.energy_system_options()
        parameters = optimization_problem.parameters(0)

        min_pressure = network_settings["pipe_minimum_pressure"]
        max_pressure = network_settings["pipe_maximum_pressure"]
        assert (
            max_pressure > min_pressure
        ), "The global maximum pressure must be larger than the minimum one."
        if np.isfinite(min_pressure) or np.isfinite(max_pressure):
            # No elevation data available yet. Assume 0 mDAT for now.
            pipe_elevation = 0.0
            min_head = min_pressure * 10.2 + pipe_elevation
            max_head = max_pressure * 10.2 + pipe_elevation
            self.__pipe_head_bounds[f"{pipe_name}.{commodity_type}In.H"] = (min_head, max_head)
            self.__pipe_head_bounds[f"{pipe_name}.{commodity_type}Out.H"] = (min_head, max_head)

        head_loss_option = network_settings["head_loss_option"]
        if head_loss_option not in HeadLossOption.__members__.values():
            raise Exception(f"Head loss option '{head_loss_option}' does not exist")

        length = parameters[f"{pipe_name}.length"]
        if length < 0.0:
            raise ValueError("Pipe length has to be larger than or equal to zero")

        if head_loss_option == HeadLossOption.NO_HEADLOSS or (
            length == 0.0 and not parameters[f"{pipe_name}.has_control_valve"]
        ):
            self.__pipe_head_loss_zero_bounds[f"{pipe_name}.dH"] = (0.0, 0.0)
            head_loss_var = ""  # Required for checking if the keys exist in the return past below
        else:
            q_nominal = self._hn_pipe_nominal_discharge(options, parameters, pipe_name)
            head_loss_nominal = self._hn_pipe_head_loss(
                pipe_name,
                optimization_problem,
                options,
                network_settings,
                parameters,
                q_nominal,
                # network_type=self.network_settings["network_type"],
                pressure=parameters[f"{pipe_name}.pressure"],
            )

            self.__pipe_head_loss_nominals[f"{pipe_name}.dH"] = head_loss_nominal

            # The .dH is by definition "Out - In". The .__head_loss is by
            # definition larger than or equal to the absolute value of dH.
            head_loss_var = f"{pipe_name}.__head_loss"

            self._hn_pipe_to_head_loss_map[pipe_name] = head_loss_var
            self.__pipe_head_loss_var[head_loss_var] = ca.MX.sym(head_loss_var)

            self.__pipe_head_loss_nominals[head_loss_var] = head_loss_nominal
            self.__pipe_head_loss_bounds[head_loss_var] = (0.0, np.inf)

            if head_loss_option == HeadLossOption.LINEARIZED_N_LINES_EQUALITY:
                # Add pipe head loss linear line segment variables
                self._pipe_linear_line_segment_map[pipe_name] = {}
                self.__pipe_linear_line_segment_var[pipe_name] = {}
                self.__pipe_linear_line_segment_var_bounds[pipe_name] = {}
                # We need to creat linear line segments for the - and + volumetric flow rate
                # possibilites. Line number 1, 2, N for the - & + side is created
                discharge_type = ["neg_discharge", "pos_discharge"]
                line_number = 0
                for dtype in discharge_type:
                    for ii_line in range(network_settings["n_linearization_lines"] * 2):
                        if ii_line < network_settings["n_linearization_lines"]:
                            dtype = discharge_type[0]
                            line_number = ii_line + 1
                        else:
                            dtype = discharge_type[1]
                            line_number = ii_line + 1 - network_settings["n_linearization_lines"]

                        # start line segment numbering from 1 up to "n_linearization_lines"
                        pipe_linear_line_segment_var_name = (
                            f"{pipe_name}__pipe_linear_line_segment_num_{line_number}_{dtype}"
                        )

                        self._pipe_linear_line_segment_map[pipe_name][
                            ii_line
                        ] = pipe_linear_line_segment_var_name
                        self.__pipe_linear_line_segment_var[pipe_name][
                            pipe_linear_line_segment_var_name
                        ] = ca.MX.sym(pipe_linear_line_segment_var_name)
                        self.__pipe_linear_line_segment_var_bounds[pipe_name][
                            pipe_linear_line_segment_var_name
                        ] = (0.0, 1.0)

        return (
            (
                self.__pipe_head_bounds[f"{pipe_name}.{commodity_type}In.H"]
                if self.__pipe_head_bounds.get(f"{pipe_name}.{commodity_type}In.H") is not None
                else self.__pipe_head_bounds
            ),
            (
                self.__pipe_head_bounds[f"{pipe_name}.{commodity_type}Out.H"]
                if self.__pipe_head_bounds.get(f"{pipe_name}.{commodity_type}Out.H") is not None
                else self.__pipe_head_bounds
            ),
            (
                self.__pipe_head_loss_zero_bounds[f"{pipe_name}.dH"]
                if self.__pipe_head_loss_zero_bounds.get(f"{pipe_name}.dH") is not None
                else self.__pipe_head_loss_zero_bounds
            ),
            (
                self._hn_pipe_to_head_loss_map[pipe_name]
                if self._hn_pipe_to_head_loss_map.get(pipe_name) is not None
                else self._hn_pipe_to_head_loss_map
            ),
            (
                self.__pipe_head_loss_var[head_loss_var]
                if self.__pipe_head_loss_var.get(head_loss_var) is not None
                else self.__pipe_head_loss_var
            ),
            (
                self.__pipe_head_loss_nominals[f"{pipe_name}.dH"]
                if self.__pipe_head_loss_nominals.get(f"{pipe_name}.dH") is not None
                else self.__pipe_head_loss_nominals
            ),
            (
                self.__pipe_head_loss_nominals[head_loss_var]
                if self.__pipe_head_loss_nominals.get(head_loss_var) is not None
                else self.__pipe_head_loss_nominals
            ),
            (
                self.__pipe_head_loss_bounds[head_loss_var]
                if self.__pipe_head_loss_bounds.get(head_loss_var) is not None
                else self.__pipe_head_loss_bounds
            ),
            (
                self._pipe_linear_line_segment_map[pipe_name]
                if self._pipe_linear_line_segment_map.get(pipe_name) is not None
                else self._pipe_linear_line_segment_map
            ),
            (
                self.__pipe_linear_line_segment_var[pipe_name]
                if self.__pipe_linear_line_segment_var.get(pipe_name) is not None
                else self.__pipe_linear_line_segment_var
            ),
            (
                self.__pipe_linear_line_segment_var_bounds[pipe_name]
                if self.__pipe_linear_line_segment_var_bounds.get(pipe_name) is not None
                else self.__pipe_linear_line_segment_var_bounds
            ),
        )

    def _hn_pipe_nominal_discharge(self, energy_system_options, parameters, pipe: str) -> float:
        """
        This function returns the nominal volumetric flow (m^3/s) through the pipe.
        """
        return parameters[f"{pipe}.area"] * energy_system_options["estimated_velocity"]

    def _hn_pipe_head_loss(
        self,
        pipe: str,
        optimization_problem,
        energy_system_options,
        network_settings,
        parameters,
        discharge: Union[ca.MX, float, np.ndarray],
        head_loss: Optional[ca.MX] = None,
        dh: Optional[ca.MX] = None,
        is_disconnected: Union[ca.MX, int] = 0,
        big_m: Optional[float] = None,
        pipe_class: Optional[PipeClass] = None,
        # network_type: NetworkSettings = NetworkSettings.NETWORK_TYPE_HEAT,
        pressure: float = 0.0,
    ) -> Union[List[Tuple[ca.MX, BT, BT]], float, np.ndarray]:
        """
        This function has two purposes:
        - return the head loss constraint expression(s) or
        - compute the head loss numerically (always positive).

        Note that there are different head loss formulations (see
        :class:`HeadLossOption`). Some formulations require the passing of
        `head_loss` (a symbol that is always positive by definition), and
        others require the passing of `dh` (which is negative when the flow is
        positive).

        When `head_loss` or `dh` is its corresponding MX symbol/expression,
        the appropriate constraint expression is returned. When `head_loss`
        and `dh` are both None, the `discharge` is assumed numerical, and the
        numerical computation of the appropriate head loss formulation is
        returned. Note that this returned numerical value is always positive,
        regardless of the sign of the discharge.

        `is_disconnected` can be used to specify whether a pipe is
        disconnected or not. This is most useful if a (boolean) ca.MX symbol
        is passed, which can then be used with a big-M formulation. The big-M
        itself then also needs to be passed via the `big_m` keyword argument.
        """

        network_type = network_settings["network_type"]

        if head_loss is None and dh is None:
            symbolic = False
        else:
            symbolic = True

        head_loss_option = network_settings["head_loss_option"]
        assert (
            head_loss_option != HeadLossOption.NO_HEADLOSS
        ), "This method should be skipped when NO_HEADLOSS is set."

        length = parameters[f"{pipe}.length"]

        if length == 0.0:
            if not symbolic:
                return np.zeros_like(discharge)
            else:
                # dH is set to zero in bounds
                return []

        if isinstance(is_disconnected, ca.MX) and not isinstance(big_m, float):
            raise ValueError("When `is_disconnected` is symbolic, `big_m` must be passed as well")
        if not symbolic and isinstance(is_disconnected, ca.MX):
            raise ValueError(
                "`is_disconnected` cannot be symbolic if the other dH/Q symbols are numeric"
            )

        if isinstance(is_disconnected, (float, int)) and is_disconnected == 1.0:
            if symbolic:
                # Pipe is always disconnected, so no head loss relationship needed
                return []
            else:
                # By definition we choose the head loss over disconnected
                # pipes to be zero.
                return 0.0

        if big_m is None:
            assert is_disconnected == 0.0
        else:
            assert big_m != 0.0

        wall_roughness = energy_system_options["wall_roughness"]
        if pipe_class is not None:
            diameter = pipe_class.inner_diameter
            area = pipe_class.area
            maximum_velocity = pipe_class.maximum_velocity
        else:
            diameter = parameters[f"{pipe}.diameter"]
            area = parameters[f"{pipe}.area"]
            maximum_velocity = network_settings["maximum_velocity"]

        try:
            # Only heat networks have a temperature attribute in the pipes, otherwise we will use
            # a default temperature for gas networks
            temperature = parameters[f"{pipe}.temperature"]
            for _id, attr in optimization_problem.temperature_carriers().items():
                if (
                    parameters[f"{pipe}.carrier_id"] == attr["id_number_mapping"]
                    and len(
                        optimization_problem.temperature_regimes(parameters[f"{pipe}.carrier_id"])
                    )
                    > 0
                ):
                    temperature = min(
                        optimization_problem.temperature_regimes(parameters[f"{pipe}.carrier_id"])
                    )
        except KeyError:
            # A default temperature of 20 degrees celcius is used for gas networks.
            temperature = 20.0

        try:
            has_control_valve = parameters[f"{pipe}.has_control_valve"]
        except KeyError:
            has_control_valve = False

        if head_loss_option == HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY:
            assert not has_control_valve

            ff = darcy_weisbach.friction_factor(
                maximum_velocity,
                diameter,
                wall_roughness,
                temperature,
                network_type=network_type,
                pressure=pressure,
            )

            # Compute c_v constant (where |dH| ~ c_v * v^2)
            c_v = length * ff / (2 * GRAVITATIONAL_CONSTANT) / diameter

            linearization_velocity = maximum_velocity
            linearization_head_loss = c_v * linearization_velocity**2
            linearization_discharge = linearization_velocity * area

            expr = linearization_head_loss * discharge / linearization_discharge

            if symbolic:
                constraint_nominal = c_v * maximum_velocity**2
                # Interior point solvers, like IPOPT, do not like linearly dependent
                # tight inequality constraints. For this reason, we split the
                # constraints depending whether the Big-M formulation is used or not.
                if big_m is None:
                    return [((-1 * dh - expr) / constraint_nominal, 0.0, 0.0)]
                else:
                    constraint_nominal = (constraint_nominal * big_m) ** 0.5

                    return [
                        (
                            (-1 * dh - expr + is_disconnected * big_m) / constraint_nominal,
                            0.0,
                            np.inf,
                        ),
                        (
                            (-1 * dh - expr - is_disconnected * big_m) / constraint_nominal,
                            -np.inf,
                            0.0,
                        ),
                    ]
            else:
                return expr * np.sign(discharge)

        elif head_loss_option in {
            HeadLossOption.CQ2_INEQUALITY,
            HeadLossOption.CQ2_EQUALITY,
        }:
            ff = darcy_weisbach.friction_factor(
                energy_system_options["estimated_velocity"],
                diameter,
                wall_roughness,
                temperature,
                network_type=network_type,
                pressure=pressure,
            )

            # Compute c_v constant (where |dH| ~ c_v * v^2)
            c_v = length * ff / (2 * GRAVITATIONAL_CONSTANT) / diameter

            v = discharge / area
            expr = c_v * v**2

            if symbolic:
                q_nominal = self.variable_nominal(f"{pipe}.Q")
                head_loss_nominal = self.variable_nominal(f"{pipe}.dH")
                constraint_nominal = (head_loss_nominal * c_v * (q_nominal / area) ** 2) ** 0.5

                if head_loss_option == HeadLossOption.CQ2_INEQUALITY:
                    ub = np.inf
                else:
                    ub = 0.0

                # Interior point solvers, like IPOPT, do not like linearly dependent
                # tight inequality constraints. For this reason, we split the
                # constraints depending whether the Big-M formulation is used or n
                if big_m is None:
                    equations = [((head_loss - expr) / constraint_nominal, 0.0, ub)]
                else:
                    constraint_nominal = (constraint_nominal * big_m) ** 0.5

                    equations = [
                        (
                            (head_loss - expr + is_disconnected * big_m) / constraint_nominal,
                            0.0,
                            np.inf,
                        )
                    ]
                    if head_loss_option == HeadLossOption.CQ2_EQUALITY:
                        equations.append(
                            (
                                (head_loss - expr - is_disconnected * big_m) / constraint_nominal,
                                -np.inf,
                                0.0,
                            ),
                        )
                return equations
            else:
                return expr

        elif (
            head_loss_option == HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
            or head_loss_option == HeadLossOption.LINEARIZED_N_LINES_EQUALITY
        ):
            n_linear_lines = network_settings["n_linearization_lines"]
            n_timesteps = len(optimization_problem.times())

            a, b = darcy_weisbach.get_linear_pipe_dh_vs_q_fit(
                diameter,
                length,
                wall_roughness,
                temperature=temperature,
                n_lines=n_linear_lines,
                v_max=maximum_velocity,
                network_type=self.network_settings["network_type"],
                pressure=parameters[f"{pipe}.pressure"],
            )

            # The function above only gives result in the positive quadrant
            # (positive head loss, positive discharge). We also need a
            # positive head loss for _negative_ discharges.
            a = np.hstack([-a, a])
            b = np.hstack([b, b])

            # Vectorize constraint for speed
            if symbolic:
                q_nominal = optimization_problem.variable_nominal(f"{pipe}.Q")
                head_loss_nominal = optimization_problem.variable_nominal(f"{pipe}.dH")
                head_loss_vec = ca.repmat(head_loss, len(a))
                discharge_vec = ca.repmat(discharge, len(a))

                if isinstance(is_disconnected, ca.MX):
                    is_disconnected_vec = ca.repmat(is_disconnected, len(a))
                else:
                    # is_disconnected_vec = is_disconnected
                    is_disconnected_vec = (
                        np.ones((len(a) * discharge.size1(), 1), dtype=float) * is_disconnected
                    )

                a_vec = np.repeat(a, discharge.size1())
                b_vec = np.repeat(b, discharge.size1())

                constraint_nominal = np.abs(head_loss_nominal * a_vec * q_nominal) ** 0.5

                if big_m is None:
                    # We write the equation such that big_m is always used, even if
                    # it is None (i.e. not used). We do have to be sure to set it to 0,
                    # because we cannot multiple with "None".
                    big_m_lin = 0.0
                else:
                    big_m_lin = big_m
                    constraint_nominal = (constraint_nominal * big_m_lin) ** 0.5

                constraints = []

                # Add weak inequality constraint, value >= 0.0 for all linear lines
                constraints.append(
                    (
                        (
                            head_loss_vec
                            - (a_vec * discharge_vec + b_vec)
                            + is_disconnected_vec * big_m_lin
                        )
                        / constraint_nominal,
                        0.0,
                        np.inf,
                    ),
                )
                if head_loss_option == HeadLossOption.LINEARIZED_N_LINES_EQUALITY:
                    # Add constraints for piece-wise linear equality

                    # Populate variable indicating if a linear line segment is active (1) or not (0)
                    # pipe_linear_line_segment: will contain variables of negative and positive
                    # discharge possibilites for the pipe. This implies if a pipe is linearized
                    # with N = 2 linear lines then pipe_linear_line_segment will have 2 * 2
                    # variables
                    # Order of linear line variables:
                    #  - negative discharge line_1, line_2
                    #  - positve discharge line_1, line_2
                    pipe_linear_line_segment = self._pipe_linear_line_segment_map[pipe]
                    is_line_segment_active = []

                    for _, ii_line_var in pipe_linear_line_segment.items():
                        # Create integer variable to activate/deactivate (1/0) a linear line
                        # segment
                        is_line_segment_active_var = optimization_problem.state_vector(ii_line_var)

                        # Linear line segment activation variable for each time step of demand
                        # profile
                        is_line_segment_active.append(is_line_segment_active_var)

                    # Calculate constraint to enforce that only 1 linear line segment can be active
                    # per time step for the current pipe for the entire time horizon
                    for itstep in range(n_timesteps):
                        is_line_segment_active_sum_per_timestep = 0.0
                        for ii_line in range(len(pipe_linear_line_segment)):

                            is_line_segment_active_sum_per_timestep = (
                                is_line_segment_active_sum_per_timestep
                                + is_line_segment_active[ii_line][itstep]
                            )
                        constraints.append(
                            (is_line_segment_active_sum_per_timestep, 1.0, 1.0),
                        )

                    # Add equality constraint, value == 0.0 for all linear lines
                    # Loop twice due to linear lines exsiting for negative and positive discharge
                    # ii==0: this is the linear lines for the negative discharge possibility
                    # ii==1: this is the linear lines for the postive discharge possibility
                    for ii in range(2):
                        for ii_line in range(n_linear_lines):
                            ii_line_used = (ii * 2) + ii_line

                            ii_start = (ii_line + 2 * ii) * n_timesteps
                            ii_end = ii_start + n_timesteps
                            constraints.append(
                                (
                                    (
                                        head_loss_vec[ii_start:ii_end]
                                        - (
                                            a_vec[ii_start:ii_end] * discharge_vec[ii_start:ii_end]
                                            + b_vec[ii_start:ii_end]
                                        )
                                        + is_disconnected_vec[ii_start:ii_end] * big_m_lin
                                        + big_m_lin
                                        * (1 - is_line_segment_active[ii_line_used][0:n_timesteps])
                                    )
                                    / constraint_nominal[ii_start:ii_end],
                                    0.0,
                                    np.inf,
                                ),
                            )
                    for ii in range(2):
                        for ii_line in range(n_linear_lines):
                            ii_line_used = (ii * 2) + ii_line

                            ii_start = (ii_line + 2 * ii) * n_timesteps
                            ii_end = ii_start + n_timesteps
                            constraints.append(
                                (
                                    (
                                        head_loss_vec[ii_start:ii_end]
                                        - (
                                            a_vec[ii_start:ii_end] * discharge_vec[ii_start:ii_end]
                                            + b_vec[ii_start:ii_end]
                                        )
                                        - is_disconnected_vec[ii_start:ii_end] * big_m_lin
                                        - big_m_lin
                                        * (1 - is_line_segment_active[ii_line_used][0:n_timesteps])
                                    )
                                    / constraint_nominal[ii_start:ii_end],
                                    -np.inf,
                                    0.0,
                                ),
                            )
                return constraints
            else:
                ret = np.amax(a * np.tile(discharge, (len(a), 1)).transpose() + b, axis=1)
                if isinstance(discharge, float):
                    ret = ret[0]
                return ret

    def _hydraulic_power(
        self,
        pipe: str,
        optimization_problem,
        energy_system_options,
        network_settings,
        parameters,
        discharge: Union[ca.MX, float, np.ndarray],
        hydraulic_power: Optional[Union[ca.MX, float, np.ndarray]] = None,
        dh: Optional[ca.MX] = None,
        is_disconnected: Union[ca.MX, int] = 0,
        big_m: Optional[float] = None,
        pipe_class: Optional[PipeClass] = None,
        flow_dir: Union[ca.MX, int] = 0,
        pressure: float = 0.0,
    ) -> Union[List[Tuple[ca.MX, BT, BT]], float, np.ndarray]:
        """
        This function has two purposes:
        - return the hydraulic power constraint expression(s) or
        - compute the hydraulic power numerically (always positive).

        Note: the discharge value can be negative. In such a case the hydraulic_power in the
        constraint expression(s) will ensure that hydraulic_power is always positive. When a
        numerical value of hydraulic power is returned, it will always be a positive value, even
        when the discharge value is negative.

        Returning linearized hydraulic power or hydraulic power constraint expressions:
        - linearized hydraulic power (1 segment):
            = max hydraulic power * (discharge / maximum discharge)
            where hydraulic power:
                = delta pressure * discharge
                = rho * g * head loss * discharge
                = rho * g * (friction factor * length * velocity**2)/(diameter * 2 * g) * discharge
        - hydraulic power constraint expressions, n_number of linear line segments:
            = b + (a * discharge vector)
            where the discharge vector represents the volumetric flow rates for each segment

        `is_disconnected` can be used to specify whether a pipe is
        disconnected or not. This is most useful if a (boolean) ca.MX symbol
        is passed, which can then be used with a big-M formulation. The big-M
        itself then also needs to be passed via the `big_m` keyword argument.
        """

        network_type = network_settings["network_type"]

        if hydraulic_power is None:
            symbolic = False
        else:
            symbolic = True

        head_loss_option = network_settings["head_loss_option"]
        assert (
            head_loss_option != HeadLossOption.NO_HEADLOSS
        ), "This method should be skipped when NO_HEADLOSS is set."

        length = parameters[f"{pipe}.length"]

        if length == 0.0:
            if not symbolic:
                return np.zeros_like(discharge)
            else:
                return []

        if isinstance(is_disconnected, ca.MX) and not isinstance(big_m, float):
            raise ValueError("When `is_disconnected` is symbolic, `big_m` must be passed as well")
        if not symbolic and isinstance(is_disconnected, ca.MX):
            raise ValueError(
                "`is_disconnected` cannot be symbolic if the hydraulic_power symbol is numeric"
            )
        if isinstance(is_disconnected, (float, int)) and is_disconnected == 1.0:
            if symbolic:
                # Pipe is always disconnected, so no hydraulic power relationship needed
                return []
            else:
                # By definition we choose the hydraulic power over disconnected
                # pipes to be zero.
                return 0.0

        if big_m is None:
            assert is_disconnected == 0.0
        else:
            assert big_m != 0.0

        wall_roughness = energy_system_options["wall_roughness"]
        if network_type == NetworkSettings.NETWORK_TYPE_HEAT:
            temperature = parameters[f"{pipe}.temperature"]
            for _id, attr in optimization_problem.temperature_carriers().items():
                if (
                    parameters[f"{pipe}.carrier_id"] == attr["id_number_mapping"]
                    and len(
                        optimization_problem.temperature_regimes(parameters[f"{pipe}.carrier_id"])
                    )
                    > 0
                ):
                    temperature = min(
                        optimization_problem.temperature_regimes(parameters[f"{pipe}.carrier_id"])
                    )
        else:
            # A default temperature of 20 degrees celcius is used for gas networks.
            temperature = 20.0
        rho = parameters[f"{pipe}.rho"]

        if pipe_class is not None:
            diameter = pipe_class.inner_diameter
            area = pipe_class.area
            maximum_velocity = pipe_class.maximum_velocity
        else:
            diameter = parameters[f"{pipe}.diameter"]
            area = parameters[f"{pipe}.area"]
            maximum_velocity = network_settings["maximum_velocity"]

        constraint_nominal = abs(
            parameters[f"{pipe}.rho"]
            * GRAVITATIONAL_CONSTANT
            * optimization_problem.variable_nominal(f"{pipe}.dH")
            * optimization_problem.variable_nominal(f"{pipe}.Q")
        )

        if head_loss_option == HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY:
            # Uitlized maximum_velocity instead of estimated_velocity (used in head loss linear
            # calc)
            ff = darcy_weisbach.friction_factor(
                maximum_velocity,
                diameter,
                wall_roughness,
                temperature,
                network_type=network_type,
                pressure=pressure,
            )
            # Compute c_k constant (where |hydraulic power| ~ c_k * v^3)
            c_k = rho * ff * length * area / 2.0 / diameter
            # Compute linearized value
            max_hydraulic_power = c_k * maximum_velocity**3
            maximum_discharge = maximum_velocity * area
            hydraulic_power_linearized = max_hydraulic_power * discharge / maximum_discharge

            if symbolic:
                if big_m is None:
                    return [
                        (
                            (hydraulic_power - hydraulic_power_linearized) / constraint_nominal,
                            0.0,
                            0.0,
                        )
                    ]
                else:
                    constraint_nominal = (constraint_nominal * big_m) ** 0.5
                    # Add constraints to enforce hydraulic_power == hydraulic_power_linearized, by
                    # via the big_m method. The value of hydraulic_power must always be a positive
                    # value. Therefore, the flow direction is taken into account for the situation
                    # when the hydraulic_power_linearized is negative (hydraulic_power_linearized =
                    # f(discharge))
                    return [
                        (
                            (
                                hydraulic_power
                                - hydraulic_power_linearized
                                + (is_disconnected + (1.0 - flow_dir)) * big_m
                            )
                            / constraint_nominal,
                            0.0,
                            np.inf,
                        ),
                        (
                            (
                                hydraulic_power
                                - hydraulic_power_linearized
                                - (is_disconnected + (1.0 - flow_dir)) * big_m
                            )
                            / constraint_nominal,
                            -np.inf,
                            0.0,
                        ),
                        (
                            (
                                hydraulic_power
                                + hydraulic_power_linearized
                                + (is_disconnected + flow_dir) * big_m
                            )
                            / constraint_nominal,
                            0.0,
                            np.inf,
                        ),
                        (
                            (
                                hydraulic_power
                                + hydraulic_power_linearized
                                - (is_disconnected + flow_dir) * big_m
                            )
                            / constraint_nominal,
                            -np.inf,
                            0.0,
                        ),
                    ]
            else:
                return abs(hydraulic_power_linearized)

        elif (
            head_loss_option == HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
            or HeadLossOption.LINEARIZED_N_LINES_EQUALITY
        ):
            n_lines = network_settings["n_linearization_lines"]
            a_coef, b_coef = darcy_weisbach.get_linear_pipe_power_hydraulic_vs_q_fit(
                rho,
                diameter,
                length,
                wall_roughness,
                temperature=temperature,
                n_lines=n_lines,
                v_max=maximum_velocity,
                network_type=self.network_settings["network_type"],
                pressure=parameters[f"{pipe}.pressure"],
            )
            discharge_vec = ca.repmat(discharge, len(a_coef))
            hydraulic_power_linearized_vec = a_coef * discharge_vec + b_coef

            if symbolic:
                hydraulic_power_vec = ca.repmat(hydraulic_power, len(a_coef))

                if isinstance(is_disconnected, ca.MX):
                    is_disconnected_vec = ca.repmat(is_disconnected, len(a_coef))
                else:
                    is_disconnected_vec = is_disconnected

                if big_m is None:
                    # We write the equation such that big_m is always used, even if
                    # it is None (i.e. not used). We do have to be sure to set it to 0,
                    # because we cannot multiple with "None".
                    big_m_lin = 0.0
                else:
                    big_m_lin = big_m
                    constraint_nominal = (constraint_nominal * big_m_lin) ** 0.5
                    # Add constraints to enforce hydraulic_power_vec >=
                    # hydraulic_power_linearized_vec. The value of hydraulic_power_vec must always
                    # be a positive value. Therefore, the flow direction is taken into account for
                    # the situation when the hydraulic_power_linearized_vec is negative
                    # (hydraulic_power_linearized_vec = f(discharge))
                return [
                    (
                        (
                            hydraulic_power_vec
                            - hydraulic_power_linearized_vec
                            + (is_disconnected_vec + (1.0 - flow_dir)) * big_m_lin
                        )
                        / constraint_nominal,
                        0.0,
                        np.inf,
                    ),
                    (
                        (
                            hydraulic_power_vec
                            + hydraulic_power_linearized_vec
                            + (is_disconnected_vec + flow_dir) * big_m_lin
                        )
                        / constraint_nominal,
                        0.0,
                        np.inf,
                    ),
                ]
            else:
                # Calc the max hydraulic power out of all the linear sections
                max_hydraulic_power_linearized = np.amax(hydraulic_power_linearized_vec, axis=1)
                if isinstance(discharge, float):
                    max_hydraulic_power_linearized = max_hydraulic_power_linearized[0]
                return abs(max_hydraulic_power_linearized)
        else:
            assert (
                head_loss_option == HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
                or head_loss_option == HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY
                or head_loss_option == HeadLossOption.LINEARIZED_N_LINES_EQUALITY
            ), "This method only caters for head_loss_option: "
            "LINEARIZED_ONE_LINE_EQUALITY & LINEARIZED_N_LINES_WEAK_INEQUALITY."

    def _pipe_head_loss_path_constraints(self, optimization_problem, _ensemble_member):
        """
        We set this constraint relating .dH to the upstream and downstream
        heads here in the Mixin for scaling purposes (dH nominal is
        calculated in pre()).
        """
        constraints = []

        if self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_HEAT:
            commodity = "Heat"
            pipe_type = "heat_pipe"
        elif self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_GAS:
            commodity = "Gas"
            pipe_type = "gas_pipe"
        else:
            raise Exception(
                f"hydraulic power can not be calculated for network type of "
                f"{self.network_settings['network_type']}"
            )

        for pipe in optimization_problem.energy_system_components.get(pipe_type, []):
            dh = optimization_problem.state(f"{pipe}.dH")
            h_down = optimization_problem.state(f"{pipe}.{commodity}Out.H")
            h_up = optimization_problem.state(f"{pipe}.{commodity}In.H")

            constraint_nominal = (
                optimization_problem.variable_nominal(f"{pipe}.dH")
                * optimization_problem.variable_nominal(f"{pipe}.{commodity}In.H")
            ) ** 0.5
            constraints.append(((dh - (h_down - h_up)) / constraint_nominal, 0.0, 0.0))

        return constraints

    def _demand_head_loss_path_constraints(self, optimization_problem, _ensemble_member):
        """
        This function adds constraints for a minimum pressure drop at demands. This minimum
        pressure drop is often required in practice to guarantee that it is possible to increase
        the flow rate at that specific demand, if needed, by opening the control valve.
        """
        constraints = []

        options = optimization_problem.energy_system_options()
        components = optimization_problem.energy_system_components

        # Convert minimum pressure at far point from bar to meter (water) head
        min_head_loss = options["minimum_pressure_far_point"] * 10.2

        for d in components.get("heat_demand", []):
            constraints.append(
                (
                    optimization_problem.state(f"{d}.HeatIn.H")
                    - optimization_problem.state(f"{d}.HeatOut.H"),
                    min_head_loss,
                    np.inf,
                )
            )

        return constraints

    def _pipe_head_loss_constraints(
        self, optimization_problem, _maximum_total_head_loss, ensemble_member
    ):
        """
        This function adds the head loss constraints for pipes. There are two options namely with
        and without pipe class optimization. In both cases we assume that disconnected pipes, pipes
        without flow have no head loss.

        Under pipe-class optimization the head loss constraints per pipe class are added and
        applied with the big_m method (is_topo_disconnected) to only activate the correct
        constraints.

        Under constant pipe class constraints for only one diameter are added.
        """
        constraints = []

        options = optimization_problem.energy_system_options()
        parameters = optimization_problem.parameters(ensemble_member)
        components = optimization_problem.energy_system_components
        # Set the head loss according to the direction in the pipes. Note that
        # the `.__head_loss` symbol is always positive by definition, but that
        # `.dH` is not (positive when flow is negative, and vice versa).
        # If the pipe is disconnected, we leave the .__head_loss symbol free
        # (and it has no physical meaning). We also do not set any discharge
        # relationship in this case (but dH is still equal to Out - In of
        # course).

        if self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_HEAT:
            commodity = "heat"
            n_type = "hn"
        elif self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_GAS:
            commodity = "gas"
            n_type = "gn"
        else:
            raise Exception(
                f"hydraulic power can not be calculated for network type of "
                f"{self.network_settings['network_type']}"
            )
        pipe_type = f"{commodity}_pipe"

        pipe_head_loss_map = getattr(optimization_problem, f"_{n_type}_pipe_to_head_loss_map")
        is_disconnected_map = getattr(optimization_problem, f"_{commodity}_pipe_disconnect_map")
        pipe_topo_class_map = getattr(
            optimization_problem, f"_{commodity}_pipe_topo_pipe_class_map"
        )
        pipe_to_flow_direct_map = getattr(
            optimization_problem, f"_{commodity}_pipe_to_flow_direct_map"
        )

        for pipe in components.get(pipe_type, []):
            if parameters[f"{pipe}.length"] == 0.0:
                # If the pipe does not have a control valve, the head loss is
                # forced to zero via bounds. If the pipe _does_ have a control
                # valve, then there still is no relationship between the
                # discharge and the head loss/dH.
                continue

            head_loss_sym = pipe_head_loss_map[pipe]

            dh = _state_vector_scaled(optimization_problem, f"{pipe}.dH", ensemble_member)
            head_loss = _state_vector_scaled(optimization_problem, head_loss_sym, ensemble_member)
            discharge = _state_vector_scaled(optimization_problem, f"{pipe}.Q", ensemble_member)

            # We need to make sure the dH is decoupled from the discharge when
            # the pipe is disconnected. Simply put, this means making the
            # below constraints trivial.
            is_disconnected_var = is_disconnected_map.get(pipe)

            if is_disconnected_var is None:
                is_disconnected = 0.0
            else:
                is_disconnected = _state_vector_scaled(
                    optimization_problem, is_disconnected_var, ensemble_member
                )

            max_discharge = None
            max_head_loss = -np.inf

            if pipe in pipe_topo_class_map:
                # Multiple diameter options for this pipe
                pipe_classes = pipe_topo_class_map[pipe]
                max_discharge = max(c.maximum_discharge for c in pipe_classes)

                for pc, pc_var_name in pipe_classes.items():
                    if pc.inner_diameter == 0.0:
                        continue

                    head_loss_max_discharge = self._hn_pipe_head_loss(
                        pipe,
                        optimization_problem,
                        options,
                        self.network_settings,
                        parameters,
                        max_discharge,
                        pipe_class=pc,
                        pressure=parameters[f"{pipe}.pressure"],
                    )

                    if self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_HEAT:
                        big_m = max(1.1 * _maximum_total_head_loss, 2 * head_loss_max_discharge)
                    elif self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_GAS:
                        big_m = max(
                            2.0 * 2.0 * _maximum_total_head_loss, 2 * head_loss_max_discharge
                        )
                    else:
                        raise RuntimeError(
                            f"{self.network_settings['network_type']} is not of type Gas or Heat "
                            f"thus no maximum total headloss is given"
                        )

                    is_topo_disconnected = 1 - optimization_problem.extra_variable(
                        pc_var_name, ensemble_member
                    )
                    is_topo_disconnected = ca.repmat(is_topo_disconnected, dh.size1())

                    # Note that we add the two booleans `is_disconnected` and
                    # `is_topo_disconnected`. This is allowed because of the way the
                    # resulting expression is used in the Big-M formulation. We only care
                    # that the expression (i.e. a single boolean or the sum of the two
                    # booleans) is either 0 when the pipe is connected, or >= 1 when it
                    # is disconnected.
                    constraints.extend(
                        self._hn_pipe_head_loss(
                            pipe,
                            optimization_problem,
                            options,
                            self.network_settings,
                            parameters,
                            discharge,
                            head_loss,
                            dh,
                            is_disconnected + is_topo_disconnected,
                            big_m,
                            pc,
                            pressure=parameters[f"{pipe}.pressure"],
                        )
                    )

                    # Contrary to the Big-M calculation above, the relation
                    # between dH and the head loss symbol requires the
                    # maximum head loss that can be realized effectively. So
                    # we pass the current pipe class's maximum discharge.
                    max_head_loss = max(
                        max_head_loss,
                        self._hn_pipe_head_loss(
                            pipe,
                            optimization_problem,
                            options,
                            self.network_settings,
                            parameters,
                            pc.maximum_discharge,
                            pipe_class=pc,
                            pressure=parameters[f"{pipe}.pressure"],
                        ),
                    )
            else:
                # Only a single diameter for this pipe. Note that we rely on
                # the diameter parameter being overridden automatically if a
                # single pipe class is set by the user.
                area = parameters[f"{pipe}.area"]
                max_discharge = self.network_settings["maximum_velocity"] * area

                is_topo_disconnected = int(parameters[f"{pipe}.diameter"] == 0.0)

                if self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_HEAT:
                    big_m = 1.1 * _maximum_total_head_loss
                elif self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_GAS:
                    big_m = 2.0 * 2.0 * _maximum_total_head_loss
                else:
                    raise RuntimeError(
                        f"{self.network_settings['network_type']} is not of type Gas or Heat thus "
                        f"no maximum total headloss is given"
                    )

                constraints.extend(
                    self._hn_pipe_head_loss(
                        pipe,
                        optimization_problem,
                        options,
                        self.network_settings,
                        parameters,
                        discharge,
                        head_loss,
                        dh,
                        is_disconnected + is_topo_disconnected,
                        big_m=big_m,
                        pressure=parameters[f"{pipe}.pressure"],
                    )
                )

                max_head_loss = self._hn_pipe_head_loss(
                    pipe,
                    optimization_problem,
                    options,
                    self.network_settings,
                    parameters,
                    max_discharge,
                    pressure=parameters[f"{pipe}.pressure"],
                )

            # Relate the head loss symbol to the pipe's dH symbol.

            # FIXME: Ugly hack. Cold pipes should be modelled completely with
            # their own integers as well.
            flow_dir = _state_vector_scaled(
                optimization_problem, pipe_to_flow_direct_map[pipe], ensemble_member
            )

            # Note that the Big-M should _at least_ cover the maximum
            # distance between `head_loss` and `dh`. If `head_loss` can be at
            # most 1.0 (= `max_head_loss`), that means our Big-M should be at
            # least double (i.e. >= 2.0). And because we do not want Big-Ms to
            # be overly tight, we include an additional factor of 2.
            big_m = 2 * 2 * max_head_loss

            constraints.append(
                (
                    (-dh - head_loss + (1 - flow_dir) * big_m) / big_m,
                    0.0,
                    np.inf,
                )
            )
            constraints.append(((dh - head_loss + flow_dir * big_m) / big_m, 0.0, np.inf))

        return constraints

    def _pipe_hydraulic_power_path_constraints(
        self, optimization_problem, _maximum_total_head_loss, ensemble_member
    ):
        """
        This function adds constraints to compute the hydraulic power that is needed to realize the
        flow, compensating the pressure drop through the pipe. Similar to the head loss constraints
        we allow two supported methods. 1) a single linear line between 0 to max velocity. 2) A
        multiple line inequality approach where one can use the minimize_head_losses == True option
        to drag down the solution to the actual physical solution.

        Note that the linearizations are made separately from the pressure drop constraints, this is
        done to avoid "stacked" overestimations.
        """
        constraints = []
        options = optimization_problem.energy_system_options()

        if self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_HEAT:
            commodity = "heat"
        elif self.network_settings["network_type"] == NetworkSettings.NETWORK_TYPE_GAS:
            commodity = "gas"
        else:
            raise Exception(
                f"hydraulic power can not be calculated for network type of "
                f"{self.network_settings['network_type']}"
            )
        pipe_type = f"{commodity}_pipe"

        pipe_topo_class_map = getattr(
            optimization_problem, f"_{commodity}_pipe_topo_pipe_class_map"
        )

        if self.network_settings["head_loss_option"] != HeadLossOption.NO_HEADLOSS:
            parameters = optimization_problem.parameters(ensemble_member)
            components = optimization_problem.energy_system_components

            for pipe in components.get(pipe_type, []):
                if parameters[f"{pipe}.length"] == 0.0:
                    # If the pipe does not have a control valve, the head loss is
                    # forced to zero via bounds. If the pipe _does_ have a control
                    # valve, then there still is no relationship between the
                    # discharge and the hydraulic_power.
                    continue

                # TODO: I think the upcomming 8 lines can be removed (e.g. head_loss_option)
                if pipe_type == "heat_pipe":
                    head_loss_option = optimization_problem._hn_get_pipe_head_loss_option(
                        pipe, self.network_settings, parameters
                    )
                else:
                    head_loss_option = self.network_settings["head_loss_option"]
                assert (
                    head_loss_option != HeadLossOption.NO_HEADLOSS
                ), "This method should be skipped when NO_HEADLOSS is set."

                discharge = optimization_problem.state(f"{pipe}.Q")
                hydraulic_power = optimization_problem.state(f"{pipe}.Hydraulic_power")
                rho = parameters[f"{pipe}.rho"]

                # 0: pipe is connected, 1: pipe is disconnected
                is_disconnected_var = getattr(
                    optimization_problem, f"_{commodity}_pipe_disconnect_map"
                ).get(pipe)
                # optimization_problem._pipe_disconnect_map.get(pipe)
                if is_disconnected_var is None:
                    is_disconnected = 0.0
                else:
                    is_disconnected = optimization_problem.state(is_disconnected_var)

                flow_dir_var = getattr(
                    optimization_problem, f"_{commodity}_pipe_to_flow_direct_map"
                ).get(pipe)
                flow_dir = optimization_problem.state(
                    flow_dir_var
                )  # 0/1: negative/positive flow direction

                if pipe in pipe_topo_class_map:
                    # Multiple diameter options for this pipe
                    pipe_classes = pipe_topo_class_map[pipe]
                    max_discharge = max(c.maximum_discharge for c in pipe_classes)
                    for pc, pc_var_name in pipe_classes.items():
                        if pc.inner_diameter == 0.0:
                            continue

                        # Calc max hydraulic power based on maximum_total_head_loss =
                        # f(max_sum_dh_pipes, max_dh_network_options)
                        max_total_hydraulic_power = 2.0 * (
                            rho * GRAVITATIONAL_CONSTANT * _maximum_total_head_loss * max_discharge
                        )

                        # is_topo_disconnected - 0: pipe selected, 1: pipe disconnected/not selected
                        # self.__pipe_topo_pipe_class_var - value 0: pipe is not selected, 1: pipe
                        # is selected
                        is_topo_disconnected = 1 - optimization_problem.variable(pc_var_name)

                        constraints.extend(
                            self._hydraulic_power(
                                pipe,
                                optimization_problem,
                                options,
                                self.network_settings,
                                parameters,
                                discharge,
                                hydraulic_power,
                                is_disconnected=is_topo_disconnected + is_disconnected,
                                big_m=max_total_hydraulic_power,
                                pipe_class=pc,
                                flow_dir=flow_dir,
                                pressure=parameters[f"{pipe}.pressure"],
                            )
                        )
                else:
                    is_topo_disconnected = int(parameters[f"{pipe}.diameter"] == 0.0)
                    max_total_hydraulic_power = 2.0 * (
                        rho
                        * GRAVITATIONAL_CONSTANT
                        * _maximum_total_head_loss
                        * parameters[f"{pipe}.area"]
                        * self.network_settings["maximum_velocity"]
                    )
                    constraints.extend(
                        self._hydraulic_power(
                            pipe,
                            optimization_problem,
                            options,
                            self.network_settings,
                            parameters,
                            discharge,
                            hydraulic_power,
                            is_disconnected=is_disconnected + is_topo_disconnected,
                            big_m=max_total_hydraulic_power,
                            flow_dir=flow_dir,
                            pressure=parameters[f"{pipe}.pressure"],
                        )
                    )
        return constraints



from numpy import nan

from .gas_base import GasTwoPort
from .._internal import BaseAsset


class Compressor(GasTwoPort, BaseAsset):
    """
    A gas compressor increases the pressure level of the flow.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "compressor"
        self.min_head = 30.0

        self.Q_nominal_in = nan
        self.Q_nominal_out = nan

        self.density_in = 2.5e3  # H2 density [g/m3] at 30bar
        self.density_out = 2.5e3  # H2 density [g/m3] at 30bar

        self.add_equation(
            (
                (self.GasIn.Q * self.density_in - self.GasOut.Q * self.density_out)
                / (self.Q_nominal_in * self.density_in * self.Q_nominal_out * self.density_out)
                ** 0.5
            )
        )



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile

import numpy as np

from rtctools.util import run_optimization_problem


class TestEndScenarioSizingAnnualized(TestCase):
    """
    Tests for end scenario sizing with annualized costs in a small network with optional assets.

    This class tests two models for a heat network: with
    and without annualized costs. It asserts the following:
    1. Under some conditions, the objective value of the annualized model is equal to the solution
    from the non annualized one.
    2. The effect of the discount rate on the objective value.
    3. The correctness of annualized capital expenditures calculations.
    4: The calculate_annuity_factor function returns the correct valuea

    """

    def test_end_scenario_sizing_annualized(self):
        from models.test_case_small_network_optional_assets_annualized.src.run_ates import (
            HeatProblem,
        )
        from models.test_case_small_network_optional_assets_annualized.src.run_annualized import (
            HeatProblemDiscAnnualizedCost,
            HeatProblemDiscAnnualizedCostModifiedParam,
            HeatProblemDiscAnnualizedCostModifiedDiscountRate,
        )
        from models.test_case_small_network_optional_assets_annualized.src import (
            run_annualized,
        )
        from mesido.financial_mixin import calculate_annuity_factor

        base_folder = Path(run_annualized.__file__).resolve().parent.parent

        solution_run_ates = run_optimization_problem(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="annualized_test_case_discount5.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        # Solution of model with annualized cost, considering a discount rate > 0
        # and a technical life > 1
        solution_annualized_cost = run_optimization_problem(
            HeatProblemDiscAnnualizedCost,
            base_folder=base_folder,
            esdl_file_name="annualized_test_case_discount5.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        # Solution of model with annualized cost, with discount rate = 0 and
        # technical life = 1 year. This model is used to compare the objective value
        # of the annualized model with the objective value of the non-discounted model.
        solution__annualized_modified_param = run_optimization_problem(
            HeatProblemDiscAnnualizedCostModifiedParam,
            base_folder=base_folder,
            esdl_file_name="annualized_test_case_discount5.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        # Solution of model with annualized cost, with discount rate = 0.
        # This model is used to test the effect of the discount rate on the objective value.
        solution__annualized_modified_discount = run_optimization_problem(
            HeatProblemDiscAnnualizedCostModifiedDiscountRate,
            base_folder=base_folder,
            esdl_file_name="annualized_test_case_discount5.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        # # Assertion 1: Model for annualized objective value with discount=0 and
        # # technical life 1 year matches the objective value of the non-discounted problem
        np.testing.assert_allclose(
            solution__annualized_modified_param.objective_value, solution_run_ates.objective_value
        )

        # Assertion 2: Undiscounted problem has a lower objective value than the discocunted one
        # due to cost of capital not considered in undiscounted problem
        np.testing.assert_array_less(
            solution__annualized_modified_discount.objective_value,
            solution_annualized_cost.objective_value,
        )

        results = solution_annualized_cost.extract_results()
        heat_producers = [1, 2]
        # Number of decimal positions for test accuracy
        decimal = 4
        for i in heat_producers:
            investment_and_installation_cost = (
                results[f"HeatProducer_{i}__investment_cost"]
                + results[f"HeatProducer_{i}__installation_cost"]
            )
            # These are the same parameters used by the model from the ESDL file:
            asset_id = solution_annualized_cost.esdl_asset_name_to_id_map[f"HeatProducer_{i}"]
            years_asset_life = solution_annualized_cost.esdl_assets[asset_id].attributes[
                "technicalLifetime"
            ]
            discount_rate = (
                solution_annualized_cost.esdl_assets[asset_id]
                .attributes["costInformation"]
                .discountRate.value
            ) / 100
            # TODO: Handle if NoneType
            discount_factor = calculate_annuity_factor(discount_rate, years_asset_life)
            annualized_capex = results[f"HeatProducer_{i}__annualized_capex"]
            # # Assertion 3: annualized capex matches the discounted investment
            # # and installation cost
            np.testing.assert_almost_equal(
                annualized_capex,
                investment_and_installation_cost * discount_factor,
                decimal,
            )

        # Assertion 4: The calculate_annuity_factor function returns the correct value
        # Checks that calculate_annuity_factor() returns 1.0 for a discount factor rate
        # of 0 and technical life of 1, and that it returns a value of 1.1 for discount
        # rate of 10%
        assert np.isclose(calculate_annuity_factor(0, 1), 1.0, atol=1e-14, rtol=0.0)
        assert np.isclose(calculate_annuity_factor(0.1, 1), 1.1, atol=1e-14, rtol=0.0)


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestEndScenarioSizingAnnualized()
    a.test_end_scenario_sizing_annualized()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



import logging

import casadi as ca

from mesido.base_component_type_mixin import BaseComponentTypeMixin
from mesido.control_variables import map_comp_type_to_control_variable
from mesido.electricity_physics_mixin import ElectricityPhysicsMixin
from mesido.gas_physics_mixin import GasPhysicsMixin
from mesido.heat_physics_mixin import HeatPhysicsMixin

import numpy as np

from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)


logger = logging.getLogger("mesido")


class PhysicsMixin(
    HeatPhysicsMixin,
    ElectricityPhysicsMixin,
    GasPhysicsMixin,
    BaseComponentTypeMixin,
    CollocatedIntegratedOptimizationProblem,
):
    """
    This class is to combine the physics of multiple commodities into a single mixin and add
    functionalities that span multiple commodities.

    Part of the physics constraints are dependent upon the asset size. What we have decided is to
    keep all logic concerning the physics in the respective PhysicsMixin, this avoids the need to
    overwrite constraints in the AssetSizingMixin with duplicate code. To allow for this principle
    to work, we do instantiate empty maps in the respective PhysicsMixins, this allows all logic for
    the constraints to be implemented in the PhysicsMixin utilizing some try/except and/or if
    statements. The AssetSizingMixin instantiates all sizing variables and the maps will be
    populated, resulting in that the correct constraints in the PhysicsMixin will be applied. This
    also enables the user to run the PhysicsMixin separately.
    """

    def __init__(self, *args, **kwargs):
        """
        In this __init__ we prepare the dicts for the variables added by the HeatMixin class
        """

        # Boolean variable for determining when a asset switches from setpoint
        self._timed_setpoints = {}
        self._change_setpoint_var = {}
        self._change_setpoint_bounds = {}
        self._component_to_change_setpoint_map = {}

        if "timed_setpoints" in kwargs and isinstance(kwargs["timed_setpoints"], dict):
            self._timed_setpoints = kwargs["timed_setpoints"]

        super().__init__(*args, **kwargs)

    def energy_system_options(self):
        r"""
        Returns a dictionary of milp network specific options.
        """

        # options = super().heat_network_options()

        options = HeatPhysicsMixin.energy_system_options(self)
        options.update(ElectricityPhysicsMixin.energy_system_options(self))
        options.update(GasPhysicsMixin.energy_system_options(self))

        return options

    def pre(self):
        """
        In this pre method we fill the dicts initiated in the __init__. This means that we create
        the Casadi variables and determine the bounds, nominals and create maps for easier
        retrieving of the variables.
        """
        super().pre()

        # Mixed-interger formulation of component setpoint
        for component_name in self._timed_setpoints.keys():
            # Make 1 variable per component (so not per control
            # variable) which represents if the setpoint of the component
            # is changed (1) is not changed (0) in a timestep
            change_setpoint_var = f"{component_name}._change_setpoint_var"
            self._component_to_change_setpoint_map[component_name] = change_setpoint_var
            self._change_setpoint_var[change_setpoint_var] = ca.MX.sym(change_setpoint_var)
            self._change_setpoint_bounds[change_setpoint_var] = (0, 1.0)

    @property
    def extra_variables(self):
        """
        In this function we add all the variables defined in the HeatMixin to the optimization
        problem. Note that these are only the normal variables not path variables.
        """
        variables = super().extra_variables.copy()

        return variables

    @property
    def path_variables(self):
        """
        In this function we add all the path variables defined in the HeatMixin to the
        optimization problem. Note that path_variables are variables that are created for each
        time-step.
        """
        variables = super().path_variables.copy()
        variables.extend(self._change_setpoint_var.values())

        return variables

    def variable_is_discrete(self, variable):
        """
        All variables that only can take integer values should be added to this function.
        """
        if variable in self._change_setpoint_var:
            return True
        else:
            return super().variable_is_discrete(variable)

    def variable_nominal(self, variable):
        """
        In this function we add all the nominals for the variables defined/added in the HeatMixin.
        """

        return super().variable_nominal(variable)

    def bounds(self):
        """
        In this function we add the bounds to the problem for all the variables defined/added in
        the HeatMixin.
        """
        bounds = super().bounds()
        bounds.update(self._change_setpoint_bounds)
        return bounds

    def __setpoint_constraint(
        self, ensemble_member, component_name, windowsize_hr, setpointchanges
    ):
        r"""Constraints that can switch only every n time steps of setpoint.
        A component can only switch setpoint every <windowsize_hr> hours.
        Apply the constraint every timestep from after the first time step onwards [from i=1].

        Inspect example curve below for understanding of dHeat/dt for
        windowsize_hr 12 with a time domain of 35 hourly timesteps.

        Heat
        d                               *-------*
        c                   *-------*
        b   *
        a       *---*---*                           *-------*


        i   0   1   2   3   4       16  17      29  30      35
        """
        assert windowsize_hr > 0
        assert windowsize_hr % 1 == 0
        assert component_name in sum(self.energy_system_components.values(), [])

        # Find the component type
        comp_type = next(
            iter(
                [
                    comptype
                    for comptype, compnames in self.energy_system_components.items()
                    for compname in compnames
                    if compname == component_name
                ]
            )
        )

        constraints = []
        times = self.times()
        control_vars = map_comp_type_to_control_variable[comp_type]
        if not isinstance(control_vars, list):
            control_vars = [control_vars]

        for var_name in control_vars:
            # Retrieve the relevant variable names
            variable_name = f"{component_name}{var_name}"
            var_name_setpoint = self._component_to_change_setpoint_map[component_name]

            # Get the timewise symbolic variables of Heat_source
            sym_var = self.__state_vector_scaled(variable_name, ensemble_member)

            # Get the timewise symbolic variables of the setpoint
            canonical, sign = self.alias_relation.canonical_signed(var_name_setpoint)
            setpoint_is_free = sign * self.state_vector(canonical, ensemble_member)

            # d<variable>/dt expression, forward Euler
            backward_heat_rate_expression = sym_var[:-1] - sym_var[1:]

            # Compute threshold for what is considered a change in setpoint
            big_m = 2.0 * max(self.bounds()[variable_name])
            nominal = self.variable_nominal(variable_name)

            # Constraint which fixes if the variable is allowed to switch or not.
            # With a dynamic sliding window, shifting one timestep.
            # Sum the binairy variables in the window. The sum should be <=1 as only on of the
            # binairy variable is allowed to represent a switch in operations.
            duration_s = 3600 * windowsize_hr

            # Number of elements to be included in a sliding window over time elements
            # Definition of elements: 23 time steps between 24 elements in the times object
            windowsize_dynamic = []
            # TODO: Better/improved/faster search function to be written
            for ij in range(0, len(times)):
                for ii in range(0 + ij, len(times)):
                    if (times[ii] - times[0 + ij]) >= duration_s:
                        windowsize_dynamic.append(ii - (0 + ij) - 1)
                        break
                else:
                    windowsize_dynamic.append(len(times) - 1 - (0 + ij))
                    break

            if len(windowsize_dynamic) == 0:
                windowsize_dynamic.append(len(times))

            for iw in range(0, len(windowsize_dynamic)):
                start_idx = iw
                if windowsize_dynamic[iw] > (len(times) - 1):
                    end_idx = len(times) - 1
                else:
                    end_idx = start_idx + windowsize_dynamic[iw]

                expression = 0.0
                for j in range(start_idx, end_idx + 1):
                    expression += setpoint_is_free[j]
                # This constraint forces that only 1 timestep in the sliding
                # window can have setpoint_is_free=1. In combination with the
                # constraints lower in this function we ensure the desired
                # behavior of limited setpoint changes.
                constraints.append(((setpointchanges - expression), 0.0, np.inf))

            # Constraints for the allowed milp rate of the component.
            # Made 2 constraints which each do or do not constrain the value
            # of the setpoint_is_free var the value of the
            # backward_heat_expression. So the discrete variable does or does
            # not have an influence on making the constrained uphold or not.

            # Note: the equations are not apply at t0

            # NOTE: we start from 2 this is to not constrain the derivative at t0
            for i in range(2, len(times)):
                # Constraining setpoint_is_free to 1 when value of
                # backward_heat_rate_expression < 0, otherwise
                # setpoint_is_free's value can be 0 and 1
                constraints.append(
                    (
                        (backward_heat_rate_expression[i - 1] + setpoint_is_free[i] * big_m)
                        / nominal,
                        0.0,
                        np.inf,
                    )
                )
                # Constraining setpoint_is_free to 1 when value of
                # backward_heat_rate_expression > 0, otherwise
                # setpoint_is_free's value can be 0 and 1
                constraints.append(
                    (
                        (backward_heat_rate_expression[i - 1] - setpoint_is_free[i] * big_m)
                        / nominal,
                        -np.inf,
                        0.0,
                    )
                )

        return constraints

    def path_constraints(self, ensemble_member):
        """
        Here we add all the path constraints to the optimization problem. Please note that the
        path constraints are the constraints that are applied to each time-step in the problem.
        """

        constraints = super().path_constraints(ensemble_member)

        return constraints

    def constraints(self, ensemble_member):
        """
        This function adds the normal constraints to the problem. Unlike the path constraints these
        are not applied to every time-step in the problem. Meaning that these constraints either
        consider global variables that are independent of time-step or that the relevant time-steps
        are indexed within the constraint formulation.
        """
        constraints = super().constraints(ensemble_member)

        for component_name, params in self._timed_setpoints.items():
            constraints.extend(
                self.__setpoint_constraint(ensemble_member, component_name, params[0], params[1])
            )

        return constraints

    def goal_programming_options(self):
        """
        Here we set the goal programming configuration. We use soft constraints for consecutive
        goals.
        """
        options = super().goal_programming_options()
        options["keep_soft_constraints"] = True
        return options

    def solver_options(self):
        """
        Here we define the solver options. By default we use the open-source solver highs and casadi
        solver qpsol.
        """
        options = super().solver_options()
        options["casadi_solver"] = "qpsol"
        options["solver"] = "highs"
        return options

    def compiler_options(self):
        """
        In this function we set the compiler configuration.
        """
        options = super().compiler_options()
        options["resolve_parameter_values"] = True
        return options

    def __state_vector_scaled(self, variable, ensemble_member):
        """
        This functions returns the casadi symbols scaled with their nominal for the entire time
        horizon.
        """
        canonical, sign = self.alias_relation.canonical_signed(variable)
        return (
            self.state_vector(canonical, ensemble_member) * self.variable_nominal(canonical) * sign
        )



from typing import Dict, List, Tuple

from mesido.heat_network_common import NodeConnectionDirection


class Topology:
    def __init__(
        self, nodes=None, gas_nodes=None, pipe_series=None, buffers=None, atess=None, busses=None
    ):
        if nodes is not None:
            self._nodes = nodes
        if gas_nodes is not None:
            self._gas_nodes = gas_nodes
        if pipe_series is not None:
            self._pipe_series = pipe_series
        if buffers is not None:
            self._buffers = buffers
        if busses is not None:
            self._busses = busses
        if atess is not None:
            self._atess = atess

    @property
    def nodes(self) -> Dict[str, Dict[int, Tuple[str, NodeConnectionDirection]]]:
        """
        Maps a node name to a dictionary of its connections. Written out using
        descriptive variable names the return type would be:
            Dict[node_name, Dict[connection_index, Tuple[connected_pipe_name, pipe_orientation]]]
        """
        try:
            return self._nodes
        except AttributeError:
            raise NotImplementedError

    @property
    def gas_nodes(self) -> Dict[str, Dict[int, Tuple[str, NodeConnectionDirection]]]:
        """
        Maps a gas_node name to a dictionary of its connections. Written out using
        descriptive variable names the return type would be:
            Dict[node_name, Dict[connection_index, Tuple[connected_pipe_name, pipe_orientation]]]
        """
        try:
            return self._gas_nodes
        except AttributeError:
            raise NotImplementedError

    @property
    def busses(self) -> Dict[str, Dict[int, Tuple[str, NodeConnectionDirection]]]:
        """
        Maps a bus name to a dictionary of its connections. Written out using
        descriptive variable names the return type would be:
            Dict[bus_name, Dict[connection_index, Tuple[connected_cable_name, cable_orientation]]]
        """
        try:
            return self._busses
        except AttributeError:
            raise NotImplementedError

    @property
    def pipe_series(self) -> List[List[str]]:
        """
        Return a list of a pipe series (which itself is a list of pipe names
        per serie). Note that all pipes in a series should have the same
        orientation.
        """
        try:
            return self._pipe_series
        except AttributeError:
            raise NotImplementedError

    @property
    def buffers(
        self,
    ) -> Dict[str, Tuple[Tuple[str, NodeConnectionDirection], Tuple[str, NodeConnectionDirection]]]:
        """
        Maps a buffer name to a dictionary of its in/out connections. Written out using
        descriptive variable names the return type would be:
            Dict[buffer_name, Tuple[Tuple[hot_pipe, hot_pipe_orientation],
                                    Tuple[cold_pipe, cold_pipe_orientation]]]
        """
        try:
            return self._buffers
        except AttributeError:
            raise NotImplementedError

    @property
    def ates(
        self,
    ) -> Dict[str, Tuple[Tuple[str, NodeConnectionDirection], Tuple[str, NodeConnectionDirection]]]:
        """
        Maps an ates name to a dictionary of its in/out connections. Written out using
        descriptive variable names the return type would be:
            Dict[buffer_name, Tuple[Tuple[hot_pipe, hot_pipe_orientation],
                                    Tuple[cold_pipe, cold_pipe_orientation]]]
        """
        try:
            return self._atess
        except AttributeError:
            raise NotImplementedError



from mesido.pycml import Variable

from numpy import nan

from .electricity_base import ElectricityPort
from .._internal import BaseAsset
from .._internal.electricity_component import ElectricityComponent


class ElectricitySource(ElectricityComponent, BaseAsset):
    """
    The electricity source component is used to generate electrical power and provide that to the
    network. As we set the equality constraint on the demand side we do not have to set any
    constraint at the source side.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "electricity_source"
        self.meret_place = 1

        self.price = nan

        self.power_nominal = nan

        self.add_variable(ElectricityPort, "ElectricityOut")  # [W]
        self.add_variable(
            Variable, "Electricity_source", min=0.0, nominal=self.power_nominal
        )  # [W]

        self.add_equation(
            ((self.ElectricityOut.Power - self.Electricity_source) / self.power_nominal)
        )



import math

import CoolProp as cP

from iapws import IAPWS95

from mesido.constants import GRAVITATIONAL_CONSTANT
from mesido.network_common import NetworkSettings

import numpy as np


def _kinematic_viscosity(temperature, network_type=NetworkSettings.NETWORK_TYPE_HEAT, pressure=0.0):
    """
    The kinematic viscosity is determined as a function of the fluid used.
    - If the network type is a milp network, the used fluid is water for which the kinematic
    viscosity barely changes with pressure, thus determined at a fixed pressure (0.5MPa) and a
    temperature [K] based on the network information.
    - If the fluid is hydrogen or gas, the kinematic viscosity is calculated using CoolProp for
    which the pressure [Pa] and temperature [K] are provided as inputs.
    The gas composition is based on Groninger gas.
    """
    if network_type == NetworkSettings.NETWORK_TYPE_HEAT:
        return IAPWS95(T=273.15 + temperature, P=0.5).nu
    elif network_type == NetworkSettings.NETWORK_TYPE_HYDROGEN:
        pressure = pressure if pressure else 101325
        return cP.CoolProp.PropsSI("V", "T", 273.15 + temperature, "P", pressure, "HYDROGEN")
    elif network_type == NetworkSettings.NETWORK_TYPE_GAS:
        pressure = pressure if pressure else 101325
        return cP.CoolProp.PropsSI(
            "V",
            "T",
            273.15 + temperature,
            "P",
            pressure,
            NetworkSettings.NETWORK_COMPOSITION_GAS,
        )
    else:
        raise Exception("Unknown network type for computing dynamic viscosity")


def _colebrook_white(reynolds, relative_roughness, friction_factor=0.015):
    """
    This function return the friction factor for turbulent conditions with the Colebrook-White
    equation.
    """
    for _ in range(1000):
        friction_factor_old = friction_factor

        reynolds_star = (
            1 / math.sqrt(8.0) * reynolds * math.sqrt(friction_factor) * relative_roughness
        )
        friction_factor = (
            1.0
            / (
                -2.0
                * math.log10(
                    2.51 / reynolds / math.sqrt(friction_factor) * (1 + reynolds_star / 3.3)
                )
            )
            ** 2
        )

        if (
            abs(friction_factor - friction_factor_old) / max(friction_factor, friction_factor_old)
            < 1e-6
        ):
            return friction_factor
    else:
        raise Exception("Colebrook-White did not converge")


def friction_factor(
    velocity,
    diameter,
    wall_roughness,
    temperature,
    network_type=NetworkSettings.NETWORK_TYPE_HEAT,
    pressure=0.0,
):
    """
    Darcy-weisbach friction factor calculation from both laminar and turbulent
    flow.
    """

    kinematic_viscosity = _kinematic_viscosity(
        temperature, network_type=network_type, pressure=pressure
    )
    reynolds = velocity * diameter / kinematic_viscosity

    assert velocity >= 0

    if velocity == 0.0 or diameter == 0.0:
        return 0.0
    elif reynolds <= 2000.0:
        friction_factor = 64.0 / reynolds
    elif reynolds >= 4000.0:
        friction_factor = _colebrook_white(reynolds, wall_roughness / diameter)
    else:
        fac_turb = _colebrook_white(4000.0, wall_roughness / diameter)
        fac_laminar = 64.0 / 2000.0
        w = (reynolds - 2000.0) / 2000.0
        friction_factor = w * fac_turb + (1 - w) * fac_laminar

    return friction_factor


def head_loss(
    velocity,
    diameter,
    length,
    wall_roughness,
    temperature,
    network_type=NetworkSettings.NETWORK_TYPE_HEAT,
    pressure=0.0,
):
    """
    Head loss for a circular pipe of given length.
    """

    f = friction_factor(
        velocity,
        diameter,
        wall_roughness,
        temperature,
        network_type=network_type,
        pressure=pressure,
    )

    return length * f / (2 * GRAVITATIONAL_CONSTANT) * velocity**2 / diameter


def get_linear_pipe_dh_vs_q_fit(
    diameter,
    length,
    wall_roughness,
    temperature,
    n_lines=10,
    v_max=2.5,
    network_type=NetworkSettings.NETWORK_TYPE_HEAT,
    pressure=0.0,
):
    """
    This function returns a set of coefficients to approximate a head loss curve with linear
    functions in the form of: head loss = b + (a * Q)
    """
    area = math.pi * diameter**2 / 4

    v_points = np.linspace(0.0, v_max, n_lines + 1)
    q_points = v_points * area

    h_points = np.array(
        [
            head_loss(
                v,
                diameter,
                length,
                wall_roughness,
                temperature,
                network_type=network_type,
                pressure=pressure,
            )
            for v in v_points
        ]
    )

    a = np.diff(h_points) / np.diff(q_points)
    b = h_points[1:] - a * q_points[1:]

    return a, b


def get_linear_pipe_power_hydraulic_vs_q_fit(
    rho,
    diameter,
    length,
    wall_roughness,
    temperature,
    n_lines=10,
    v_max=2.5,
    network_type=NetworkSettings.NETWORK_TYPE_HEAT,
    pressure=0.0,
):
    """
    power_hydraulic = b + (a * Q)
    """
    area = math.pi * diameter**2 / 4.0

    v_points = np.linspace(0.0, v_max, n_lines + 1)
    q_points = v_points * area
    power_hydraulic_points = np.array(
        [
            rho
            * GRAVITATIONAL_CONSTANT
            * abs(
                head_loss(
                    v,
                    diameter,
                    length,
                    wall_roughness,
                    temperature,
                    network_type=network_type,
                    pressure=pressure,
                )
            )
            * v
            * area
            for v in v_points
        ]
    )

    a = np.diff(power_hydraulic_points) / np.diff(q_points)  # calc gradients for n_line segments
    b = power_hydraulic_points[1:] - a * q_points[1:]

    return a, b



from .. import ElectricitySource


class SolarPV(ElectricitySource):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_subtype = "solar_pv"
